import itertools
import sys
import uuid

import pytest
import requests

from globus_sdk import ErrorSubdocument, GlobusAPIError, exc
from globus_sdk.testing import construct_error


def _strmatch_any_order(inputstr, prefix, midfixes, suffix, sep=", "):
    # test for string matching, but without insisting on ordering of middle segments
    assert inputstr in [
        prefix + sep.join(m) + suffix for m in itertools.permutations(midfixes)
    ]


def test_raw_json_property():
    data = {
        "code": "Json Error",
        "errors": [
            {
                "message": "json error message",
                "title": "json error title",
            }
        ],
    }

    err = construct_error(body=data, http_status=400)
    assert err.raw_json == data


@pytest.mark.parametrize(
    "body, response_headers",
    (
        ("text_data", {}),  # plain text
        ("{", {"Content-Type": "application/json"}),  # malformed JSON
    ),
)
def test_raw_json_none_on_nonjson_data(body, response_headers):
    err = construct_error(body=body, response_headers=response_headers, http_status=400)
    assert err.raw_json is None


def test_text_property():
    err = construct_error(body="text_data", http_status=400)
    assert err.text == "text_data"


def test_binary_content_property():
    body_text = "some data"
    err = construct_error(body=body_text, http_status=400)
    assert err.binary_content == body_text.encode("utf-8")


# `add_note()` and `__notes__` are new in Python 3.11
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="Exception.add_note() is new in Python 3.11"
)
def test_notes_are_populated_with_text():
    text_body = f"some error: {uuid.uuid4()}"
    err = construct_error(body=text_body, http_status=400)
    assert err.text == text_body
    assert any(text_body in note for note in err.__notes__)


@pytest.mark.parametrize(
    "body, response_headers, http_status, expect_code, expect_message",
    (
        ("text_data", {}, 401, None, "Unauthorized"),  # text
        # JSON with unrecognized contents
        (
            {"foo": "bar"},
            {"Content-Type": "application/json"},
            403,
            None,
            "Forbidden",
        ),
        # JSON with well-known contents
        (
            {"code": "foo", "message": "bar"},
            {"Content-Type": "application/json"},
            403,
            "foo",
            "bar",
        ),
        # non-dict JSON with well-known contents
        (
            "[]",
            {"Content-Type": "application/json"},
            403,
            None,
            "Forbidden",
        ),
        # invalid JSON
        (
            "{",
            {"Content-Type": "application/json"},
            400,
            None,
            "Bad Request",
        ),
    ),
)
def test_get_args(body, response_headers, http_status, expect_code, expect_message):
    err = construct_error(
        body=body, response_headers=response_headers, http_status=http_status
    )
    req = err._underlying_response.request
    assert err._get_args() == [
        req.method,
        req.url,
        None,
        http_status,
        expect_code,
        expect_message,
    ]


def test_info_is_falsey_on_non_dict_json():
    err = construct_error(
        body="[]",
        response_headers={"Content-Type": "application/json"},
        http_status=400,
    )
    assert bool(err.info.consent_required) is False
    assert bool(err.info.authorization_parameters) is False
    assert str(err.info) == "AuthorizationParameterInfo(:)|ConsentRequiredInfo(:)"


@pytest.mark.parametrize(
    "body, is_detected, required_scopes",
    (
        (
            {"code": "ConsentRequired", "required_scopes": ["foo", "bar"]},
            True,
            ["foo", "bar"],
        ),
        (
            {"code": "ConsentRequired", "required_scope": "foo bar"},
            True,
            ["foo bar"],
        ),
        ({"code": "ConsentRequired"}, False, None),
        ({"code": "ConsentRequired", "required_scopes": []}, False, None),
        ({"code": "ConsentRequired", "required_scopes": ["foo", 123]}, False, None),
        ({"code": "ConsentRequired", "required_scope": 1}, False, None),
    ),
)
def test_consent_required_info(body, is_detected, required_scopes):
    err = construct_error(body=body, http_status=403)

    if is_detected:
        assert bool(err.info.consent_required) is True
        assert err.info.consent_required.required_scopes == required_scopes
    else:
        assert bool(err.info.consent_required) is False


def test_consent_required_info_str():
    info = exc.ConsentRequiredInfo(
        {"code": "ConsentRequired", "required_scopes": ["foo", "bar"]}
    )
    assert str(info) == "ConsentRequiredInfo(required_scopes=['foo', 'bar'])"

    info = exc.ConsentRequiredInfo({})
    assert str(info) == "ConsentRequiredInfo(:)"


def test_authz_params_info_containing_session_message():
    body = {"authorization_parameters": {"session_message": "foo"}}
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message == "foo"
    assert err.info.authorization_parameters.session_required_identities is None
    assert err.info.authorization_parameters.session_required_single_domain is None
    assert err.info.authorization_parameters.session_required_policies is None
    print("derk")
    print(str(err.info.authorization_parameters))
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=foo",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


def test_authz_params_info_containing_malformed_session_message():
    body = {"authorization_parameters": {"session_message": 100}}
    err = construct_error(error_class=GlobusAPIError, body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message is None
    assert err.info.authorization_parameters.session_required_identities is None
    assert err.info.authorization_parameters.session_required_single_domain is None
    assert err.info.authorization_parameters.session_required_policies is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


def test_authz_params_info_containing_session_required_identities():
    body = {"authorization_parameters": {"session_required_identities": ["foo", "bar"]}}
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message is None
    assert err.info.authorization_parameters.session_required_identities == [
        "foo",
        "bar",
    ]
    assert err.info.authorization_parameters.session_required_single_domain is None
    assert err.info.authorization_parameters.session_required_policies is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=['foo', 'bar']",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


def test_authz_params_info_containing_malformed_session_required_identities():
    body = {"authorization_parameters": {"session_required_identities": "foo,bar"}}
    err = construct_error(error_class=GlobusAPIError, body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message is None
    assert err.info.authorization_parameters.session_required_identities is None
    assert err.info.authorization_parameters.session_required_single_domain is None
    assert err.info.authorization_parameters.session_required_policies is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


def test_authz_params_info_containing_session_required_single_domain():
    body = {
        "authorization_parameters": {"session_required_single_domain": ["foo", "bar"]}
    }
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message is None
    assert err.info.authorization_parameters.session_required_identities is None
    assert err.info.authorization_parameters.session_required_single_domain == [
        "foo",
        "bar",
    ]
    assert err.info.authorization_parameters.session_required_policies is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=['foo', 'bar']",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


def test_authz_params_info_containing_malformed_session_required_single_domain():
    body = {"authorization_parameters": {"session_required_single_domain": "foo,bar"}}
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message is None
    assert err.info.authorization_parameters.session_required_identities is None
    assert err.info.authorization_parameters.session_required_single_domain is None
    assert err.info.authorization_parameters.session_required_policies is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


@pytest.mark.parametrize("policies_value", ["foo,bar", ["foo", "bar"]])
def test_authz_params_info_containing_session_required_policies(policies_value):
    body = {"authorization_parameters": {"session_required_policies": policies_value}}
    err = construct_error(error_class=GlobusAPIError, body=body, http_status=403)
    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_message is None
    assert err.info.authorization_parameters.session_required_identities is None
    assert err.info.authorization_parameters.session_required_single_domain is None
    assert err.info.authorization_parameters.session_required_policies == ["foo", "bar"]
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=['foo', 'bar']",
            "session_required_mfa=None",
        ],
        ")",
    )


def test_authz_params_info_containing_malformed_session_required_policies():
    # confirm that if `session_required_policies` is not str|list[str],
    # it will parse as `None`
    body = {"authorization_parameters": {"session_required_policies": {"foo": "bar"}}}
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_required_policies is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


@pytest.mark.parametrize("mfa_value", [True, False])
def test_authz_params_info_containing_session_required_mfa(mfa_value):
    body = {"authorization_parameters": {"session_required_mfa": mfa_value}}
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_required_mfa is mfa_value
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            f"session_required_mfa={mfa_value}",
        ],
        ")",
    )


def test_authz_params_info_containing_malformed_session_required_mfa():
    body = {"authorization_parameters": {"session_required_mfa": "foobarjohn"}}
    err = construct_error(body=body, http_status=403)

    assert bool(err.info.authorization_parameters) is True
    assert err.info.authorization_parameters.session_required_mfa is None
    _strmatch_any_order(
        str(err.info.authorization_parameters),
        "AuthorizationParameterInfo(",
        [
            "session_message=None",
            "session_required_identities=None",
            "session_required_single_domain=None",
            "session_required_policies=None",
            "session_required_mfa=None",
        ],
        ")",
    )


@pytest.mark.parametrize(
    "orig, wrap_class",
    [
        (requests.RequestException("exc_message"), exc.NetworkError),
        (requests.Timeout("timeout_message"), exc.GlobusTimeoutError),
        (
            requests.ConnectTimeout("connect_timeout_message"),
            exc.GlobusConnectionTimeoutError,
        ),
        (requests.ConnectionError("connection_message"), exc.GlobusConnectionError),
    ],
)
def test_requests_err_wrappers(orig, wrap_class):
    msg = "dummy message"
    err = wrap_class(msg, orig)
    assert err.underlying_exception == orig
    assert str(err) == msg


@pytest.mark.parametrize(
    "orig, conv_class",
    [
        (requests.RequestException("exc_message"), exc.NetworkError),
        (requests.Timeout("timeout_message"), exc.GlobusTimeoutError),
        (
            requests.ConnectTimeout("connect_timeout_message"),
            exc.GlobusConnectionTimeoutError,
        ),
        (requests.ConnectionError("connection_message"), exc.GlobusConnectionError),
    ],
)
def test_convert_requests_exception(orig, conv_class):
    conv = exc.convert_request_exception(orig)
    assert conv.underlying_exception == orig
    assert isinstance(conv, conv_class)


@pytest.mark.parametrize(
    "status, expect_reason",
    [
        (400, "Bad Request"),
        (500, "Internal Server Error"),
    ],
)
def test_http_reason_exposure(status, expect_reason):
    body = {"errors": [{"message": "json error message", "code": "Json Error"}]}
    err = construct_error(body=body, http_status=status)
    assert err.http_reason == expect_reason


def test_http_header_exposure(make_response):
    body = {"errors": [{"message": "json error message", "code": "Json Error"}]}
    err = construct_error(
        body=body,
        http_status=400,
        response_headers={"Content-Type": "application/json", "Spam": "Eggs"},
    )
    assert err.headers["Spam"] == "Eggs"
    assert err.headers["Content-Type"] == "application/json"


# do not parametrize each of these independently: it would result in hundreds of tests
# which are not meaningfully non-overlapping in what they test
# instead, iterate through "full variations" to keep the suite faster
@pytest.mark.parametrize(
    "http_method, http_status, error_code, request_url, error_message, authz_scheme",
    [
        (
            "POST",
            404,
            "FooError",
            "https://bogus.example.com/foo",
            "got a foo error",
            "bearer",
        ),
        ("PATCH", 500, None, "https://bogus.example.org/bar", "", "unknown-token"),
        ("PUT", 501, None, "https://bogus.example.org/bar", "Not Implemented", None),
    ],
)
def test_error_repr_has_expected_info(
    http_method,
    http_status,
    authz_scheme,
    request_url,
    error_code,
    error_message,
):
    http_reason = {404: "Not Found", 500: "Server Error", 501: "Not Implemented"}.get(
        http_status
    )

    body = {"otherfield": "otherdata"}
    if error_code is not None:
        body["code"] = error_code
    if error_message is not None:
        body["message"] = error_message

    headers = {"Content-Type": "application/json", "Spam": "Eggs"}
    request_headers = {}
    if authz_scheme is not None:
        request_headers["Authorization"] = f"{authz_scheme} TOKENINFO"

    # build the response -> error -> error repr
    err = construct_error(
        body=body,
        http_status=http_status,
        method=http_method,
        url=request_url,
        response_headers=headers,
        request_headers=request_headers,
    )
    stringified = repr(err)

    # check using substring -- do not check exact format
    assert http_method in stringified
    assert request_url in stringified
    if authz_scheme in GlobusAPIError.RECOGNIZED_AUTHZ_SCHEMES:
        assert authz_scheme in stringified
    # confirm that actual tokens don't get into the repr, regardless of authz scheme
    assert "TOKENINFO" not in stringified
    assert str(http_status) in stringified
    if error_code is not None:
        assert error_code in stringified
    else:
        # several things could be 'None', but at least one of them is 'code'
        assert "None" in stringified
    if error_message is None:
        assert "otherdata" in stringified
    else:
        assert "otherdata" not in stringified
        if error_message:
            assert error_message in stringified
        else:
            assert http_reason in stringified


@pytest.mark.parametrize(
    "content_type",
    ("application/json", "application/unknown+json", "application/vnd.api+json"),
)
def test_loads_jsonapi_error_subdocuments(content_type):
    body = {
        "errors": [
            {
                "code": "TooShort",
                "title": "Password data too short",
                "detail": "password was only 3 chars long, must be at least 8",
            },
            {
                "code": "MissingSpecial",
                "title": "Password data missing special chars",
                "detail": "password must have non-alphanumeric characters",
            },
            {
                "code": "ContainsCommonDogName",
                "title": "Password data has a popular dog name",
                "detail": "password cannot contain 'spot', 'woofy', or 'clifford'",
            },
        ]
    }
    err = construct_error(
        body=body, http_status=422, response_headers={"Content-Type": content_type}
    )

    # code is not taken from any of the subdocuments (inherently too ambiguous)
    # this holds regardless of which parsing path was taken
    assert err.code is None

    # messages can be extracted, and they prefer detail to title
    assert err.messages == [
        "password was only 3 chars long, must be at least 8",
        "password must have non-alphanumeric characters",
        "password cannot contain 'spot', 'woofy', or 'clifford'",
    ]


@pytest.mark.parametrize(
    "content_type",
    ("application/json", "application/unknown+json", "application/vnd.api+json"),
)
def test_loads_jsonapi_error_subdocuments_with_common_code(content_type):
    body = {
        "errors": [
            {
                "code": "MissingClass",
                "title": "Must contain capital letter",
                "detail": "password must contain at least one capital letter",
            },
            {
                "code": "MissingClass",
                "title": "Must contain special chars",
                "detail": "password must have non-alphanumeric characters",
            },
        ]
    }
    err = construct_error(
        body=body, http_status=422, response_headers={"Content-Type": content_type}
    )
    # code is taken because all subdocuments have the same code
    assert err.code == "MissingClass"


@pytest.mark.parametrize(
    "content_type",
    ("application/json", "application/unknown+json", "application/vnd.api+json"),
)
def test_loads_jsonapi_error_messages_from_various_fields(content_type):
    body = {
        "errors": [
            {
                "message": "invalid password value",
            },
            {
                "title": "Must contain capital letter",
            },
            {
                "detail": "password must have non-alphanumeric characters",
            },
        ]
    }
    err = construct_error(
        body=body, http_status=422, response_headers={"Content-Type": content_type}
    )

    # no code was found
    assert err.code is None

    # messages are extracted, and they use whichever field is appropriate for
    # each sub-error
    # note that 'message' will *not* be extracted if the Content-Type indicated JSON:API
    # because JSON:API does not define such a field
    if content_type.endswith("vnd.api+json"):
        assert err.messages == [
            "Must contain capital letter",
            "password must have non-alphanumeric characters",
        ]
    else:
        assert err.messages == [
            "invalid password value",
            "Must contain capital letter",
            "password must have non-alphanumeric characters",
        ]


@pytest.mark.parametrize(
    "error_doc",
    (
        # Type Zero Error Format
        {"code": "FooCode", "message": "FooMessage"},
        # Undefined Error Format
        {"message": "FooMessage"},
    ),
)
def test_non_jsonapi_parsing_uses_root_as_errors_array_by_default(error_doc):
    err = construct_error(body=error_doc, http_status=422)

    # errors is the doc root wrapped in a list, but converted to a subdocument error
    assert len(err.errors) == 1
    subdoc = err.errors[0]
    assert isinstance(subdoc, ErrorSubdocument)
    assert subdoc.raw == error_doc
    # note that 'message' is supported for error message extraction
    # vs 'detail' and 'title' for JSON:API data
    assert subdoc.message == error_doc["message"]


@pytest.mark.parametrize(
    "error_doc",
    (
        # Type Zero Error Format with sub-error data
        {"code": "FooCode", "message": "FooMessage", "errors": [{"bar": "baz"}]},
        # Type Zero Error Format with *empty* sub-error data
        {"code": "FooCode", "message": "FooMessage", "errors": []},
        # Undefined Error Format with sub-error data
        {"message": "FooMessage", "errors": [{"bar": "baz"}]},
        # Undefined Error Format with *empty* sub-error data
        {"message": "FooMessage", "errors": []},
    ),
)
def test_non_jsonapi_parsing_uses_errors_array_if_present(error_doc):
    err = construct_error(body=error_doc, http_status=422)

    # errors is the 'errors' list converted to error subdocs
    # first some sanity checks...
    assert len(err.errors) == len(error_doc["errors"])
    assert all(isinstance(subdoc, ErrorSubdocument) for subdoc in err.errors)
    # ...and then a true equivalence test
    assert [e.raw for e in err.errors] == error_doc["errors"]



import http
import sys
import typing as t

import pytest

from globus_sdk.testing import RegisteredResponse


@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="test requires http.HTTPMethod (Python 3.11+)"
)
def test_registered_response_method_literal_type_is_correct():
    all_known_methods = [m.value for m in http.HTTPMethod]
    init_signature = t.get_type_hints(RegisteredResponse.__init__)

    method_arg_type = init_signature["method"]
    expected_method_arg_type = t.Literal[tuple(all_known_methods)]

    assert method_arg_type == expected_method_arg_type



import json

from globus_sdk.experimental import BookmarkUpdateDocument
from globus_sdk.testing import get_last_request, load_response


def test_update_bookmark(client):
    meta = load_response(client.update_bookmark).metadata

    data = BookmarkUpdateDocument(
        meta["name"],
        meta["path"],
        pinned=meta["pinned"],
    )

    res = client.update_bookmark(meta["bookmark_id"], data)

    assert res.http_status == 200
    assert res["data"]["type"] == "Bookmark"
    assert (
        res["data"]["relationships"]["collection"]["data"]["id"] == meta["collection"]
    )

    req = get_last_request()
    sent = json.loads(req.body)
    assert sent["data"]["attributes"]["name"] == meta["name"]
    assert sent["data"]["attributes"]["path"] == meta["path"]
    assert sent["data"]["attributes"]["pinned"] == meta["pinned"]



from unittest import mock

import pytest

import globus_sdk

REFRESH_TOKEN = "refresh_token_1"
ACCESS_TOKEN = "access_token_1"
EXPIRES_AT = -1


@pytest.fixture(params=["simple", "with_new_refresh_token"])
def response(request):
    r = mock.Mock()
    r.by_resource_server = {
        "simple": {"rs1": {"expires_at_seconds": -1, "access_token": "access_token_2"}},
        "with_new_refresh_token": {
            "rs1": {
                "expires_at_seconds": -1,
                "access_token": "access_token_2",
                "refresh_token": "refresh_token_2",
            }
        },
    }[request.param]
    return r


@pytest.fixture
def client(response):
    c = mock.Mock()
    c.oauth2_refresh_token = mock.Mock(return_value=response)
    return c


@pytest.fixture
def authorizer(client):
    return globus_sdk.RefreshTokenAuthorizer(
        REFRESH_TOKEN, client, access_token=ACCESS_TOKEN, expires_at=EXPIRES_AT
    )


def test_get_token_response(authorizer, client, response):
    """
    Calls _get_token_response, confirms that the mock
    AuthClient is used and the known data was returned.
    """
    # get new_access_token
    res = authorizer._get_token_response()
    assert res == response
    # confirm mock ConfidentialAppAuthClient was used as expected
    client.oauth2_refresh_token.assert_called_once_with(REFRESH_TOKEN)


def test_multiple_resource_servers(authorizer, response):
    """
    Sets the mock client to return multiple resource servers.
    Confirms GlobusError is raised when _extract_token_data is called.
    """
    response.by_resource_server["rs2"] = {
        "expires_at_seconds": -1,
        "access_token": "access_token_3",
    }
    with pytest.raises(ValueError) as excinfo:
        authorizer._extract_token_data(response)
    assert "didn't return exactly one token" in str(excinfo.value)


def test_conditional_refresh_token_update(authorizer, response):
    """
    Call ensure_valid_token (triggering a refresh)
    Confirm that the authorizer always updates its access token and only updates
    refresh_token if one was present in the response
    """
    authorizer.ensure_valid_token()  # trigger refresh
    token_data = response.by_resource_server["rs1"]
    if "refresh_token" in token_data:  # if present, confirm refresh token was updated
        assert authorizer.access_token == "access_token_2"
        assert authorizer.refresh_token == "refresh_token_2"
    else:  # otherwise, confirm no change
        assert authorizer.access_token == "access_token_2"
        assert authorizer.refresh_token == "refresh_token_1"


def test_refresh_token_authorizer_rejects_auth_service_client():
    with pytest.raises(
        globus_sdk.GlobusSDKUsageError,
        match="RefreshTokenAuthorizer requires an AuthLoginClient",
    ):
        globus_sdk.RefreshTokenAuthorizer(REFRESH_TOKEN, globus_sdk.AuthClient())



from __future__ import annotations

import typing as t
from urllib.parse import parse_qs, urlsplit

from .base import PageT, Paginator


class JSONAPIPaginator(Paginator[PageT]):
    """
    A paginator which uses the next link of a JSON:API response if present
    to fetch more pages.

    Assumes the next link only changes query parameters between calls to
    `method` and that `method` supports a keyword argument of `query_params`
    which those changed query parameters will be passed to.
    """

    _REQUIRES_METHOD_KWARGS = ("query_params",)

    def _get_next_link(self, page: dict[str, t.Any]) -> str | None:
        links = page.get("links")
        if links:
            next_link = links.get("next")
            if isinstance(next_link, str):
                return next_link
        return None

    def pages(self) -> t.Iterator[PageT]:
        while True:
            current_page = self.method(*self.client_args, **self.client_kwargs)
            yield current_page

            next_link = self._get_next_link(current_page)
            if next_link:
                next_link_query_params = parse_qs(urlsplit(next_link).query)
                self.client_kwargs["query_params"] = next_link_query_params
            else:
                break



import urllib.parse

import pytest

import globus_sdk
from globus_sdk import MISSING
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize(
    "include_param",
    [MISSING, "private_policies", "private_policies,foo", ("private_policies", "foo")],
)
def test_get_storage_gateway_list(client, include_param):
    meta = load_response(client.get_storage_gateway_list).metadata
    expect_ids = meta["ids"]

    res = client.get_storage_gateway_list(include=include_param)
    assert res.http_status == 200

    # confirm iterable and sanity check some fields
    assert len(list(res)) > 0
    for sg in res:
        assert sg["DATA_TYPE"] == "storage_gateway#1.0.0"
        assert "id" in sg
        assert "display_name" in sg
    assert [sg["id"] for sg in res] == expect_ids

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    if include_param is MISSING:
        assert parsed_qs == {}
    elif isinstance(include_param, str):
        assert parsed_qs == {"include": [include_param]}
    else:
        assert parsed_qs == {"include": [",".join(include_param)]}


def test_create_storage_gateway(client):
    meta = load_response(client.create_storage_gateway).metadata

    # the SDK does not validate the create document, so an empty document is fine, if
    # unrealistic, with a mocked response
    res = client.create_storage_gateway({})
    assert res.http_status == 200

    # confirm top level access to storage gateway data
    assert res["id"] == meta["id"]
    assert res["display_name"] == meta["display_name"]


def test_create_storage_gateway_validation_error(client):
    meta = load_response(
        client.create_storage_gateway, case="validation_error"
    ).metadata

    with pytest.raises(globus_sdk.GCSAPIError) as excinfo:
        client.create_storage_gateway({})

    error = excinfo.value
    assert error.http_status == meta["http_status"]
    assert error.code == meta["code"]
    assert error.message == meta["message"]


@pytest.mark.parametrize(
    "include_param",
    [MISSING, "private_policies", "private_policies,foo", ("private_policies", "foo")],
)
def test_get_storage_gateway(client, include_param):
    meta = load_response(client.get_storage_gateway).metadata

    res = client.get_storage_gateway(meta["id"], include=include_param)
    assert res.http_status == 200

    # confirm top level access to storage gateway data
    assert res["id"] == meta["id"]
    assert res["display_name"] == meta["display_name"]

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    if include_param is MISSING:
        assert parsed_qs == {}
    elif isinstance(include_param, str):
        assert parsed_qs == {"include": [include_param]}
    else:
        assert parsed_qs == {"include": [",".join(include_param)]}


def test_update_storage_gateway(client):
    meta = load_response(client.update_storage_gateway).metadata

    # as in the create test, an empty update document is not very realistic
    # but because there's no request validation, this is fine
    res = client.update_storage_gateway(meta["id"], {})
    assert res.http_status == 200

    # confirm top level access to response data
    assert res["code"] == "success"
    assert res["message"] == "Operation successful"


def test_delete_storage_gateway(client):
    meta = load_response(client.delete_storage_gateway).metadata

    res = client.delete_storage_gateway(meta["id"])
    assert res.http_status == 200

    # confirm top level access to response data
    assert res["code"] == "success"
    assert res["message"] == "Operation successful"


def test_delete_storage_gateway_permission_denied(client):
    meta = load_response(
        client.delete_storage_gateway, case="permission_denied_error"
    ).metadata

    with pytest.raises(globus_sdk.GCSAPIError) as excinfo:
        client.delete_storage_gateway(meta["id"])

    error = excinfo.value
    assert error.http_status == meta["http_status"]
    assert error.code == meta["code"]
    assert error.message == meta["message"]



from .clients import GetClientsResponse
from .consents import GetConsentsResponse
from .credentials import GetClientCredentialsResponse
from .identities import GetIdentitiesResponse, GetIdentityProvidersResponse
from .oauth import (
    OAuthAuthorizationCodeResponse,
    OAuthClientCredentialsResponse,
    OAuthDependentTokenResponse,
    OAuthRefreshTokenResponse,
    OAuthTokenResponse,
)
from .policies import GetPoliciesResponse
from .projects import GetProjectsResponse
from .scopes import GetScopesResponse

__all__ = (
    "GetClientCredentialsResponse",
    "GetClientsResponse",
    "GetIdentitiesResponse",
    "GetIdentityProvidersResponse",
    "GetConsentsResponse",
    "GetPoliciesResponse",
    "GetProjectsResponse",
    "GetScopesResponse",
    "OAuthAuthorizationCodeResponse",
    "OAuthClientCredentialsResponse",
    "OAuthDependentTokenResponse",
    "OAuthRefreshTokenResponse",
    "OAuthTokenResponse",
)



import pytest

import globus_sdk
from globus_sdk.testing import construct_error


def test_construct_error_defaults_to_base_error_class():
    err = construct_error(body="foo", http_status=400)
    assert isinstance(err, globus_sdk.GlobusAPIError)


@pytest.mark.parametrize(
    "error_class",
    (
        globus_sdk.SearchAPIError,
        globus_sdk.AuthAPIError,
        globus_sdk.TransferAPIError,
        globus_sdk.FlowsAPIError,
        globus_sdk.GCSAPIError,
    ),
)
def test_construct_error_can_be_customized_to_service_error_classes(error_class):
    err = construct_error(body="foo", http_status=400, error_class=error_class)
    assert isinstance(err, error_class)


def test_construct_error_defaults_to_json_for_dict_body():
    err = construct_error(body={"foo": "bar"}, http_status=400)
    assert err.text == '{"foo": "bar"}'
    assert err.headers == {"Content-Type": "application/json"}


@pytest.mark.parametrize(
    "body, add_params, expect_encoding, expect_text",
    (
        (b"foo-bar", {}, "utf-8", "foo-bar"),
        (b"foo-bar", {"response_encoding": "utf-8"}, "utf-8", "foo-bar"),
        (b"foo-bar", {"response_encoding": "latin-1"}, "latin-1", "foo-bar"),
        # this is invalid utf-8 (continuation byte),
        # but valid in latin-1 (e with acute accent)
        (b"\xe9", {"response_encoding": "latin-1"}, "latin-1", "é"),
    ),
)
def test_construct_error_allows_binary_content(
    body, add_params, expect_encoding, expect_text
):
    err = construct_error(body=body, http_status=400, **add_params)
    assert err.binary_content == body
    assert err.text == expect_text
    assert err._underlying_response.encoding == expect_encoding



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

TASK_ID = str(uuid.uuid4())

RESPONSES = ResponseSet(
    metadata={"task_id": TASK_ID},
    default=RegisteredResponse(
        service="transfer",
        path=f"/v0.10/task/{TASK_ID}/cancel",
        method="POST",
        json={
            "DATA_TYPE": "result",
            "code": "Canceled",
            "message": "The task has been cancelled successfully.",
            "resource": f"/task/{TASK_ID}/cancel",
            "request_id": "ABCdef789",
        },
    ),
)



#!/usr/bin/env python

import argparse
import os

import globus_sdk
from globus_sdk.token_storage import SimpleJSONFileAdapter

MY_FILE_ADAPTER = SimpleJSONFileAdapter(
    os.path.expanduser("~/.sdk-manage-projects.json")
)

SCOPES = [globus_sdk.AuthClient.scopes.manage_projects, "openid", "email"]
RESOURCE_SERVER = globus_sdk.AuthClient.resource_server

# tutorial client ID
# we recommend replacing this with your own client for any production use-cases
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

NATIVE_CLIENT = globus_sdk.NativeAppAuthClient(CLIENT_ID)


def do_login_flow(*, session_params: dict | None = None):
    NATIVE_CLIENT.oauth2_start_flow(requested_scopes=SCOPES)
    # special note!
    # this works because oauth2_get_authorize_url supports session error data
    # as parameters to build the authorization URL
    # you could do this manually with the following supported parameters:
    #   - session_required_identities
    #   - session_required_single_domain
    #   - session_required_policies
    authorize_url = NATIVE_CLIENT.oauth2_get_authorize_url(**session_params)
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")
    auth_code = input("Please enter the code here: ").strip()
    tokens = NATIVE_CLIENT.oauth2_exchange_code_for_tokens(auth_code)
    return tokens


def get_tokens():
    if not MY_FILE_ADAPTER.file_exists():
        # do a login flow, getting back initial tokens
        response = do_login_flow()
        # now store the tokens and pull out the correct token
        MY_FILE_ADAPTER.store(response)
        tokens = response.by_resource_server[RESOURCE_SERVER]
    else:
        # otherwise, we already did login; load the tokens from that file
        tokens = MY_FILE_ADAPTER.get_token_data(RESOURCE_SERVER)

    return tokens


def get_auth_client():
    tokens = get_tokens()
    return globus_sdk.AuthClient(
        authorizer=globus_sdk.AccessTokenAuthorizer(tokens["access_token"])
    )


def create_project(args):
    auth_client = get_auth_client()
    userinfo = auth_client.userinfo()
    print(
        auth_client.create_project(
            args.name,
            contact_email=userinfo["email"],
            admin_ids=userinfo["sub"],
        )
    )


def delete_project(args):
    auth_client = get_auth_client()
    print(auth_client.delete_project(args.project_id))


def list_projects():
    auth_client = get_auth_client()
    for project in auth_client.get_projects():
        print(f"name: {project['display_name']}")
        print(f"id: {project['id']}")
        print()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("action", choices=["create", "delete", "list"])
    parser.add_argument("-p", "--project-id", help="Project ID for delete")
    parser.add_argument("-n", "--name", help="Project name for create")
    args = parser.parse_args()

    try:
        execute(parser, args)
    except globus_sdk.GlobusAPIError as err:
        if not err.info.authorization_parameters:
            raise

        err_params = err.info.authorization_parameters
        session_params = {}
        if err_params.session_required_identities:
            print("session required identities detected")
            session_params["session_required_identities"] = (
                err_params.session_required_identities
            )
        if err_params.session_required_single_domain:
            print("session required single domain detected")
            session_params["session_required_single_domain"] = (
                err_params.session_required_single_domain
            )
        if err_params.session_required_policies:
            print("session required policies detected")
            session_params["session_required_policies"] = (
                err_params.session_required_policies
            )
        print(session_params)
        print(err_params)
        response = do_login_flow(session_params=session_params)
        # now store the tokens
        MY_FILE_ADAPTER.store(response)
        print(
            "Reauthenticated successfully to satisfy "
            "session requirements. Will now try again.\n"
        )

        # try the action again
        execute(parser, args)


def execute(parser, args):
    if args.action == "create":
        if args.name is None:
            parser.error("create requires --name")
        create_project(args)
    elif args.action == "delete":
        if args.project_id is None:
            parser.error("delete requires --project-id")
        delete_project(args)
    elif args.action == "list":
        list_projects()
    else:
        raise NotImplementedError()


if __name__ == "__main__":
    main()



"""
Classes used by the LocalServerLoginFlowManager to automatically receive an
auth code from a redirect after user authentication through a locally run web-server.

These classes generally shouldn't need to be used directly.
"""

from __future__ import annotations

import importlib.resources
import os
import queue
import socket
import sys
import time
import typing as t
from datetime import timedelta
from http.server import BaseHTTPRequestHandler, HTTPServer
from string import Template
from urllib.parse import parse_qsl, urlparse

from . import html_files
from .errors import LocalServerLoginError

_IS_WINDOWS = os.name == "nt"

DEFAULT_HTML_TEMPLATE = Template(
    importlib.resources.files(html_files)
    .joinpath("local_server_landing_page.html")
    .read_text()
)


class RedirectHandler(BaseHTTPRequestHandler):
    """
    BaseHTTPRequestHandler to be used by RedirectHTTPServer.
    Displays the RedirectHTTPServer's html_template and parses auth_code out of
    the redirect url.
    """

    server: RedirectHTTPServer

    def do_GET(self) -> None:
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

        html_template = self.server.html_template
        query_params = dict(parse_qsl(urlparse(self.path).query))
        code = query_params.get("code")
        if code:
            self.wfile.write(
                html_template.substitute(
                    post_login_message="", login_result="Login successful"
                ).encode("utf-8")
            )
            self.server.return_code(code)
        else:
            msg = query_params.get("error_description", query_params.get("error"))

            self.wfile.write(
                html_template.substitute(
                    post_login_message=msg, login_result="Login failed"
                ).encode("utf-8")
            )

            self.server.return_code(LocalServerLoginError(msg))


class RedirectHTTPServer(HTTPServer):
    """
    An HTTPServer which accepts an HTML `Template` to be displayed to the user
    and uses a Queue to receive an auth_code from its RequestHandler.
    """

    WAIT_TIMEOUT = timedelta(minutes=15)

    def __init__(
        self,
        server_address: tuple[str, int],
        handler_class: type[BaseHTTPRequestHandler],
        html_template: Template,
    ) -> None:
        super().__init__(server_address, handler_class)

        self.html_template = html_template
        self._auth_code_queue: queue.Queue[str | BaseException] = queue.Queue()

    def handle_error(
        self,
        request: socket.socket | tuple[bytes, socket.socket],
        client_address: t.Any,
    ) -> None:
        _, excval, _ = sys.exc_info()
        assert excval is not None
        self._auth_code_queue.put(excval)

    def return_code(self, code: str | BaseException) -> None:
        self._auth_code_queue.put_nowait(code)

    def wait_for_code(self) -> str | BaseException:
        # Windows needs special handling as blocking prevents ctrl-c interrupts
        if _IS_WINDOWS:
            deadline = time.time() + self.WAIT_TIMEOUT.total_seconds()
            while time.time() < deadline:
                try:
                    return self._auth_code_queue.get()
                except queue.Empty:
                    time.sleep(1)
        else:
            try:
                return self._auth_code_queue.get(block=True, timeout=3600)
            except queue.Empty:
                pass
        raise LocalServerLoginError("Login timed out. Please try again.")



from ._auth_requirements_error import GARE, GlobusAuthorizationParameters
from ._functional_api import has_gares, is_gare, to_gare, to_gares

__all__ = (
    "GARE",
    "GlobusAuthorizationParameters",
    "to_gare",
    "to_gares",
    "is_gare",
    "has_gares",
)



import pytest


@pytest.fixture
def process_signature(sphinxext):
    def func(signature, return_annotation):
        return sphinxext.after_autodoc_signature_replace_MISSING_repr(
            object(), "", "", object(), None, signature, return_annotation
        )

    return func


def test_autodoc_signature_MISSING_hook_skips_nulls(process_signature):
    assert process_signature(None, None) == (None, None)


def test_autodoc_signature_MISSING_hook_can_update_signature(process_signature):
    input_sig = "(x: str | <globus_sdk.MISSING> = <globus_sdk.MISSING>, y: int)"
    output_sig, output_annotation = process_signature(input_sig, "str")
    assert output_annotation == "str"
    assert output_sig == "(x: str | MISSING = MISSING, y: int)"


def test_autodoc_signature_MISSING_hook_can_update_return_type(process_signature):
    input_sig = "(x: str | int = 0, y: int = 0)"
    output_sig, output_annotation = process_signature(
        input_sig, "int | <globus_sdk.MISSING>"
    )
    assert input_sig == output_sig
    assert output_annotation == "int | MISSING"



"""
Tests for IterableTransferResponse responses from TransferClient
"""

import globus_sdk
from tests.common import register_api_route

SERVER_LIST_TEXT = """{
  "DATA": [
    {
      "DATA_TYPE": "server",
      "hostname": "ep1.transfer.globus.org",
      "id": 207976,
      "incoming_data_port_end": null,
      "incoming_data_port_start": null,
      "is_connected": true,
      "is_paused": false,
      "outgoing_data_port_end": null,
      "outgoing_data_port_start": null,
      "port": 2811,
      "scheme": "gsiftp",
      "subject": null,
      "uri": "gsiftp://ep1.transfer.globus.org:2811"
    }
  ],
  "DATA_TYPE": "endpoint_server_list",
  "endpoint": "go#ep1"
}"""


def test_server_list(client):
    epid = "epid"
    register_api_route(
        "transfer", f"/endpoint/{epid}/server_list", body=SERVER_LIST_TEXT
    )

    res = client.endpoint_server_list(epid)
    # it should still be a subclass of GlobusHTTPResponse
    assert isinstance(res, globus_sdk.GlobusHTTPResponse)

    # fetch top-level attrs
    assert res["DATA_TYPE"] == "endpoint_server_list"
    assert res["endpoint"] == "go#ep1"

    # intentionally access twice -- unlike PaginatedResource, this is allowed
    # and works
    assert len(list(res)) == 1
    assert len(list(res)) == 1

    assert list(res)[0]["DATA_TYPE"] == "server"



from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk import client, paging, response
from globus_sdk.response import IterableJSONAPIResponse
from globus_sdk.scopes import TransferScopes
from globus_sdk.services.transfer.errors import TransferAPIError
from globus_sdk.transport import RetryConfig

from .data import (
    BookmarkCreateDocument,
    BookmarkUpdateDocument,
    TunnelCreateDocument,
    TunnelUpdateDocument,
)
from .transport import TRANSFER_V2_DEFAULT_RETRY_CHECKS

log = logging.getLogger(__name__)


class TransferClientV2(client.BaseClient):
    r"""
    Client for the `/v2/ routes of the Globus Transfer API
    <https://docs.globus.org/api/transfer/#v2>`_.

    .. sdk-sphinx-copy-params:: BaseClient

    This class provides helper methods for /v2/ functionality in the Transfer
    API as it is developed.

    **Paginated Calls**

    Methods which support pagination can be called as paginated or unpaginated methods.
    If the method name is ``TransferClientV2.foo``, the paginated version is
    ``TransferClientV2.paginated.foo``.
    Using ``TransferClientV2.list_stream_access_points`` as an example::

        from globus_sdk.experimental import TransferClientV2
        tc = TransferClientV2(...)

        # this is the unpaginated version
        for stream_access_point in tc.list_stream_access_points():
            print(stream_access_point["attributes"]["display_name"])

        # this is the paginated version
        for page in tc.paginated.list_stream_access_points():
            for stream_access_point in page:
                print(stream_access_point["attributes"]["display_name"])

    .. automethodlist:: globus_sdk.experimental.TransferClientV2
    """

    service_name = "transfer"
    error_class = TransferAPIError
    scopes = TransferScopes
    default_scope_requirements = [TransferScopes.all]

    def _register_standard_retry_checks(self, retry_config: RetryConfig) -> None:
        """Override the default retry checks."""
        retry_config.checks.register_many_checks(TRANSFER_V2_DEFAULT_RETRY_CHECKS)

    # Tunnel methods

    def create_tunnel(
        self,
        data: dict[str, t.Any] | TunnelCreateDocument,
    ) -> response.GlobusHTTPResponse:
        """
        :param data: Parameters for the tunnel creation

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    data = globus_sdk.experimental.TunnelCreateDocument(...)
                    result = tc.create_tunnel(data)
                    print(result["data"]["id"])

            .. tab-item:: API Info

                ``POST /v2/tunnels``
        """
        log.debug(f"{self.__class__.__name__}.create_tunnel(...)")
        r = self.post("/v2/tunnels", data=data)
        return r

    def update_tunnel(
        self,
        tunnel_id: str | uuid.UUID,
        update_doc: dict[str, t.Any] | TunnelUpdateDocument,
    ) -> response.GlobusHTTPResponse:
        r"""
        :param tunnel_id: The ID of the Tunnel.
        :param update_doc: The document that will be sent to the patch API.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    data = globus_sdk.experimental.TunnelUpdateDocument(...)
                    result = tc.update_tunnel(tunnel_id, data)
                    print(result["data"])

            .. tab-item:: API Info

                ``PATCH /v2/tunnels/<tunnel_id>``
        """
        log.debug(f"{self.__class__.__name__}.update_tunnel({tunnel_id}, {update_doc})")
        r = self.patch(f"/v2/tunnels/{tunnel_id}", data=update_doc)
        return r

    def get_tunnel(
        self,
        tunnel_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param tunnel_id: The ID of the Tunnel which we are fetching details about.
        :param query_params: Any additional parameters will be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    result = tc.get_tunnel(tunnel_id)
                    print(result["data"])

            .. tab-item:: API Info

                ``GET /v2/tunnels/<tunnel_id>``
        """
        log.debug(f"{self.__class__.__name__}.get_tunnel({tunnel_id}, {query_params})")
        r = self.get(f"/v2/tunnels/{tunnel_id}", query_params=query_params)
        return r

    def delete_tunnel(
        self,
        tunnel_id: str | uuid.UUID,
    ) -> response.GlobusHTTPResponse:
        """
        :param tunnel_id: The ID of the Tunnel to be deleted.

        This will clean up all data associated with a Tunnel.
        Note that Tunnels must be stopped before they can be deleted.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    tc.delete_tunnel(tunnel_id)

            .. tab-item:: API Info

                ``DELETE /v2/tunnels/<tunnel_id>``
        """
        log.debug(f"{self.__class__.__name__}.delete_tunnel({tunnel_id})")
        r = self.delete(f"/v2/tunnels/{tunnel_id}")
        return r

    def list_tunnels(
        self,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableJSONAPIResponse:
        """
        :param query_params: Any additional parameters will be passed through
            as query params.

        This will list all the Tunnels created by the authorized user.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    tc.list_tunnels(tunnel_id)

            .. tab-item:: API Info

                ``GET /v2/tunnels/``
        """
        log.debug(f"{self.__class__.__name__}.list_tunnels({query_params})")
        r = self.get("/v2/tunnels", query_params=query_params)
        return IterableJSONAPIResponse(r)

    def get_tunnel_events(
        self,
        tunnel_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableJSONAPIResponse:
        """
        :param tunnel_id: The ID of the Tunnel which we are fetching events about.
        :param query_params: Any additional parameters will be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    result = tc.get_tunnel_events(tunnel_id)
                    print(result["data"])

            .. tab-item:: API Info

                ``GET /v2/tunnels/<tunnel_id>/events``
        """
        log.debug(
            f"{self.__class__.__name__}.get_tunnel_events({tunnel_id}, {query_params})"
        )
        r = self.get(f"/v2/tunnels/{tunnel_id}/events", query_params=query_params)
        return IterableJSONAPIResponse(r)

    # Stream access point methods

    def get_stream_access_point(
        self,
        stream_ap_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param stream_ap_id: The ID of the steaming access point to lookup.
        :param query_params: Any additional parameters will be passed through
            as query params.

        Get a stream access point by id.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    tc.get_stream_access_point(stream_ap_id)

            .. tab-item:: API Info

                ``GET /v2/stream_access_points/<stream_ap_id>``
        """
        log.debug(
            f"{self.__class__.__name__}."
            f"get_stream_access_point({stream_ap_id}, {query_params})"
        )
        r = self.get(
            f"/v2/stream_access_points/{stream_ap_id}", query_params=query_params
        )
        return r

    @paging.has_paginator(paging.JSONAPIPaginator)
    def list_stream_access_points(
        self,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableJSONAPIResponse:
        """
        :param query_params: Any additional parameters will be passed through
            as query params.

        List stream access points.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    tc.list_stream_access_points()

            .. tab-item:: API Info

                ``GET /v2/stream_access_points``
        """
        log.debug(
            f"{self.__class__.__name__}.list_stream_access_points({query_params})"
        )
        r = self.get("/v2/stream_access_points", query_params=query_params)
        return IterableJSONAPIResponse(r)

    def create_bookmark(
        self, data: dict[str, t.Any] | BookmarkCreateDocument
    ) -> response.GlobusHTTPResponse:
        """
        :param data: Parameters for bookmark creation

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    data = globus_sdk.experimental.BookmarkCreateDocument(...)
                    result = tc.create_bookmark(data)
                    print(result["data"]["id"])

            .. tab-item:: API Info

                ``POST /v2/bookmarks``
        """
        log.debug(f"{self.__class__.__name__}.create_bookmark(...)")
        r = self.post("/v2/bookmarks", data=data)
        return r

    def update_bookmark(
        self,
        bookmark_id: uuid.UUID | str,
        update_document: dict[str, t.Any] | BookmarkUpdateDocument,
    ) -> response.GlobusHTTPResponse:
        r"""
        :param bookmark_id: The ID of the Bookmark.
        :param update_document: The document that will be sent to the patch API.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    data = globus_sdk.experimental.BookmarkUpdateDocument(...)
                    result = tc.update_bookmark(bookmark_id, data)
                    print(result["data"])

            .. tab-item:: API Info

                ``PATCH /v2/bookmarks/<bookmark_id>``
        """
        log.debug(
            f"{self.__class__.__name__}."
            f"update_bookmark({bookmark_id}, {update_document})"
        )
        r = self.patch(f"/v2/bookmarks/{bookmark_id}", data=update_document)
        return r

    def get_bookmark(
        self,
        bookmark_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param bookmark_id: The ID of the Bookmark for which we are fetching details.
        :param query_params: Any additional parameters will be passed through
            as request parameters on the URL.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    query_params = {"include": "collection"}
                    result = tc.get_bookmark(bookmark_id, query_params)
                    print(result["data"])

            .. tab-item:: API Info

                ``GET /v2/bookmarks/<bookmark_id>``
        """
        log.debug(
            f"{self.__class__.__name__}.get_bookmark({bookmark_id}, {query_params})"
        )
        r = self.get(f"/v2/bookmarks/{bookmark_id}", query_params=query_params)
        return r

    def delete_bookmark(
        self,
        bookmark_id: uuid.UUID | str,
    ) -> response.GlobusHTTPResponse:
        """
        :param bookmark_id: The ID of the Bookmark to be deleted.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    tc.delete_bookmark(bookmark_id)

            .. tab-item:: API Info

                ``DELETE /v2/bookmarks/<bookmark_id>``
        """
        log.debug(f"{self.__class__.__name__}.delete_bookmark({bookmark_id})")
        r = self.delete(f"/v2/bookmarks/{bookmark_id}")
        return r

    def list_bookmarks(
        self,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableJSONAPIResponse:
        """
        :param query_params: Any additional parameters will be passed through
            as request parameters on the URL.

        This will list all Bookmarks created by the authenticated user's primary
            and linked identities.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.experimental.TransferClientV2(...)
                    query_params = {"include": "collection"}
                    tc.list_bookmarks(query_params)

            .. tab-item:: API Info

                ``GET /v2/bookmarks``
        """
        log.debug(f"{self.__class__.__name__}.list_bookmarks({query_params})")
        r = self.get("/v2/bookmarks", query_params=query_params)
        return IterableJSONAPIResponse(r)



from __future__ import annotations

import typing as t
import uuid

from .client import AuthClient


def is_username(val: str) -> bool:
    # If the value parses as a UUID, then it's an ID, not a username.
    # If it does not parse as such, then it must be a username.
    try:
        uuid.UUID(val)
        return False
    except ValueError:
        return True


def split_ids_and_usernames(
    identity_ids: t.Iterable[str],
) -> tuple[set[str], set[str]]:
    ids = set()
    usernames = set()

    for val in identity_ids:
        if is_username(val):
            usernames.add(val)
        else:
            ids.add(val)

    return ids, usernames


class IdentityMap:
    r"""
    There's a common pattern of having a large batch of Globus Auth Identities which you
    want to inspect. For example, you may have a list of identity IDs fetched from
    Access Control Lists on Globus Endpoints. In order to display these identities to an
    end user, you may want to resolve them to usernames.

    However, naively looking up the identities one-by-one is very inefficient. It's best
    to do batched lookups with multiple identities at once. In these cases, an
    ``IdentityMap`` can be used to do those batched lookups for you.

    An ``IdentityMap`` is a mapping-like type which converts Identity IDs and Identity
    Names to Identity records (dictionaries) using the Globus Auth API.

    .. note::

        ``IdentityMap`` objects are not full Mappings in the same sense as python dicts
        and similar objects. By design, they only implement a small part of the Mapping
        protocol.

    The basic usage pattern is

    - create an ``IdentityMap`` with an AuthClient which will be used to call out to
      Globus Auth

    - seed the ``IdentityMap`` with IDs and Usernames via :py:meth:`~IdentityMap.add` (you
      can also do this during initialization)

    - retrieve identity IDs or Usernames from the map

    Because the map can be populated with a collection of identity IDs and Usernames
    prior to lookups being performed, it can improve the efficiency of these operations
    up to 100x over individual lookups.

    If you attempt to retrieve an identity which has not been previously added to the
    map, it will be immediately added. But adding many identities beforehand will
    improve performance.

    The ``IdentityMap`` will cache its results so that repeated lookups of the same Identity
    will not repeat work. It will also map identities both by ID and by Username,
    regardless of how they're initially looked up.

    .. warning::

        If an Identity is not found in Globus Auth, it will trigger a KeyError when
        looked up. Your code must be ready to handle KeyErrors when doing a lookup.

    Correct usage looks something like so::

        ac = globus_sdk.AuthClient(...)
        idmap = globus_sdk.IdentityMap(
            ac, ["foo@globusid.org", "bar@uchicago.edu"]
        )
        idmap.add("baz@xsede.org")
        # adding by ID is also valid
        idmap.add("c699d42e-d274-11e5-bf75-1fc5bf53bb24")
        # map ID to username
        assert (
            idmap["c699d42e-d274-11e5-bf75-1fc5bf53bb24"]["username"]
            == "go@globusid.org"
        )
        # map username to ID
        assert (
            idmap["go@globusid.org"]["id"]
            == "c699d42e-d274-11e5-bf75-1fc5bf53bb24"
        )

    And simple handling of errors::

        try:
            record = idmap["no-such-valid-id@example.org"]
        except KeyError:
            username = "NO_SUCH_IDENTITY"
        else:
            username = record["username"]

    or you may achieve this by using the :py:meth:`~.IdentityMap.get` method::

        # internally handles the KeyError and returns the default value
        record = idmap.get("no-such-valid-id@example.org", None)
        username = record["username"] if record is not None else "NO_SUCH_IDENTITY"

    :param auth_client: The client object which will be used for lookups against Globus Auth
    :param identity_ids: A list or other iterable of usernames or identity IDs (potentially
        mixed together) which will be used to seed the ``IdentityMap`` 's tracking of
        unresolved Identities.
    :param id_batch_size: A non-default batch size to use when communicating with Globus
        Auth. Leaving this set to the default is strongly recommended.
    :param cache:  A dict or other mapping object which will be used to cache results.
        The default is that results are cached once per IdentityMap object. If you want
        multiple IdentityMaps to share data, explicitly pass the same ``cache`` to both.

    .. automethodlist:: globus_sdk.IdentityMap
        :include_methods: __getitem__,__delitem__
    """  # noqa

    _default_id_batch_size = 100

    def __init__(
        self,
        auth_client: AuthClient,
        identity_ids: t.Iterable[str] | None = None,
        *,
        id_batch_size: int | None = None,
        cache: None | t.MutableMapping[str, dict[str, t.Any]] = None,
    ) -> None:
        self.auth_client = auth_client
        self.id_batch_size = id_batch_size or self._default_id_batch_size

        # uniquify, copy, and split into IDs vs usernames
        self.unresolved_ids, self.unresolved_usernames = split_ids_and_usernames(
            [] if identity_ids is None else identity_ids
        )

        # a cache may be passed in via the constructor in order to make multiple
        # IdentityMap objects share a cache
        self._cache = cache if cache is not None else {}

    def _create_batch(self, key: str) -> set[str]:
        """
        Create a batch to do a lookup.

        For whichever set of unresolved names is appropriate, build the batch to
        lookup up to *at most* the batch size. Also, remove the unresolved names from
        tracking so that they will not be looked up again.
        """
        key_is_username = is_username(key)
        set_to_use = (
            self.unresolved_usernames if key_is_username else self.unresolved_ids
        )

        # start the batch with the key being looked up, and if it is in the unresolved
        # list remove it
        batch = {key}
        if key in set_to_use:
            set_to_use.remove(key)

        # until we've exhausted the set or filled the batch, keep trying to add
        while set_to_use and len(batch) < self.id_batch_size:
            value = set_to_use.pop()

            # value may already have been looked up if the cache is shared, skip those
            if value in self._cache:
                continue

            batch.add(value)

        return batch

    def _fetch_batch_including(self, key: str) -> None:
        """
        Batch resolve identifiers (usernames or IDs), being sure to include the desired,
        named key. The key also determines which kind of batch will be built --
        usernames or IDs.

        Store the results in the internal cache.
        """
        batch = self._create_batch(key)

        if is_username(key):
            response = self.auth_client.get_identities(usernames=batch)
        else:
            response = self.auth_client.get_identities(ids=batch)

        for x in response["identities"]:
            self._cache[x["id"]] = x
            self._cache[x["username"]] = x

    def add(self, identity_id: str) -> bool:
        """
        Add a username or ID to the ``IdentityMap`` for batch lookups later.

        Returns True if the ID was added for lookup.
        Returns False if it was rejected as a duplicate of an already known name.

        :param identity_id: A string Identity ID or Identity Name (a.k.a. "username") to
            add
        """
        if identity_id in self._cache:
            return False
        if is_username(identity_id):
            if identity_id in self.unresolved_usernames:
                return False
            else:
                self.unresolved_usernames.add(identity_id)
                return True
        if identity_id in self.unresolved_ids:
            return False
        self.unresolved_ids.add(identity_id)
        return True

    def get(self, key: str, default: t.Any | None = None) -> t.Any:
        """
        A dict-like get() method which accepts a default value.

        :param key: The username or ID to look up
        :param default: The default value to return if the key is not found
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: str) -> t.Any:
        """
        ``IdentityMap`` supports dict-like lookups with ``map[key]``
        """
        if key not in self._cache:
            self._fetch_batch_including(key)
        return self._cache[key]

    def __delitem__(self, key: str) -> None:
        """
        ``IdentityMap`` supports ``del map[key]``. Note that this only removes lookup
        values from the cache and will not impact the set of unresolved/pending IDs.
        """
        del self._cache[key]



from __future__ import annotations

import importlib
import re
import typing as t

import responses

import globus_sdk
import globus_sdk.experimental

from .models import RegisteredResponse, ResponseList, ResponseSet

# matches "V2", "V11", etc as a string suffix
# see usage in _resolve_qualname for details
_SUFFIX_VERSION_MATCH_PATTERN = re.compile(r"V\d+$")

_RESPONSE_SET_REGISTRY: dict[t.Any, ResponseSet] = {}


def register_response_set(
    set_id: t.Any,
    rset: ResponseSet | dict[str, dict[str, t.Any]],
    metadata: dict[str, t.Any] | None = None,
) -> ResponseSet:
    """
    Register a new ``ResponseSet`` object.

    The response set may be specified as a dict or a ResponseSet object.

    :param set_id: The ID used to retrieve the response set later
    :param rset: The response set to register
    :param metadata: Metadata dict to assign to the response set when it is specified
        as a dict. If the response set is an object, this argument is ignored.
    """
    if isinstance(rset, dict):
        rset = ResponseSet.from_dict(rset, metadata=metadata)
    _RESPONSE_SET_REGISTRY[set_id] = rset
    return rset


def _resolve_qualname(name: str) -> str:
    if "." not in name:
        return name
    prefix, suffix = name.split(".", 1)

    # something from globus_sdk, could be a client class
    if hasattr(globus_sdk, prefix):
        maybe_client = getattr(globus_sdk, prefix)
    elif hasattr(globus_sdk.experimental, prefix):
        maybe_client = getattr(globus_sdk.experimental, prefix)
    else:
        return name

    # there are a dozen ways of writing this check, but the point is
    # "if it's not a client class"
    if not (
        isinstance(maybe_client, type)
        and issubclass(maybe_client, globus_sdk.BaseClient)
    ):
        return name

    assert issubclass(maybe_client, globus_sdk.BaseClient)
    service_name = maybe_client.service_name

    # TODO: Consider alternative strategies for mapping versioned clients
    # to subdirs. For now, we do it by name matching.
    #
    # 'prefix' is the client name, and it may end in `V2`, `V3`, etc.
    # in which case we want to map it to a subdir
    suffix_version_match = _SUFFIX_VERSION_MATCH_PATTERN.search(prefix)
    if suffix_version_match:
        suffix = f"{suffix_version_match.group(0).lower()}.{suffix}"

    return f"{service_name}.{suffix}"


def get_response_set(set_id: t.Any) -> ResponseSet:
    """
    Lookup a ``ResponseSet`` as in ``load_response_set``, but without
    activating it.

    :param set_id: The ID used to retrieve the response set. Typically a string, but
        could be any key used to register a response set.
    """
    # first priority: check the explicit registry
    if set_id in _RESPONSE_SET_REGISTRY:
        return _RESPONSE_SET_REGISTRY[set_id]

    # if ID is a string, it's the (optionally dotted) name of a module
    if isinstance(set_id, str):
        module_name = f"globus_sdk.testing.data.{set_id}"
    else:
        assert hasattr(
            set_id, "__qualname__"
        ), f"cannot load response set from {type(set_id)}"
        # support modules like
        #   globus_sdk/testing/data/auth/get_identities.py
        # for lookups like
        #   get_response_set(AuthClient.get_identities)
        module_name = (
            f"globus_sdk.testing.data.{_resolve_qualname(set_id.__qualname__)}"
        )

    # after that, check the built-in "registry" built from modules
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        raise ValueError(f"no fixtures defined for {module_name}") from e
    assert isinstance(module.RESPONSES, ResponseSet)
    return module.RESPONSES


def load_response_set(
    set_id: t.Any, *, requests_mock: responses.RequestsMock | None = None
) -> ResponseSet:
    """
    Optionally lookup a response set and activate all of its responses. If
    passed a ``ResponseSet``, activate it, otherwise the first argument is an
    ID used for lookup.

    :param set_id: The ID used to retrieve the response set. Typically a string, but
        could be any key used to register a response set.
    :param requests_mock: A ``responses`` library mock to use for response mocking,
        defaults to the ``responses`` default
    """
    if isinstance(set_id, ResponseSet):
        return set_id.activate_all(requests_mock=requests_mock)
    ret = get_response_set(set_id)
    ret.activate_all(requests_mock=requests_mock)
    return ret


def load_response(
    set_id: t.Any,
    *,
    case: str = "default",
    requests_mock: responses.RequestsMock | None = None,
) -> RegisteredResponse | ResponseList:
    """
    Optionally lookup and activate an individual response. If given a
    ``RegisteredResponse``, activate it, otherwise the first argument is an ID
    of a ``ResponseSet`` used for lookup. By default, looks for the response
    registered under ``case="default"``.

    :param set_id: The ID used to retrieve the response set. Typically a string, but
        could be any key used to register a response set.
    :param case: The name of a case within the response set to load, ignoring all other
        registered mocks in the response set
    :param requests_mock: A ``responses`` library mock to use for response mocking,
        defaults to the ``responses`` default
    """
    if isinstance(set_id, RegisteredResponse):
        return set_id.add(requests_mock=requests_mock)
    rset = get_response_set(set_id)
    return rset.activate(case, requests_mock=requests_mock)



from unittest import mock

import pytest
import responses

import globus_sdk


@pytest.fixture(autouse=True)
def mocksleep():
    with mock.patch("time.sleep") as m:
        yield m


@pytest.fixture(autouse=True)
def mocked_responses():
    """
    All tests enable `responses` patching of the `requests` package, replacing
    all HTTP calls.
    """
    responses.start()

    yield

    responses.stop()
    responses.reset()


@pytest.fixture
def make_response():
    def _make_response(
        response_class=None,
        status=200,
        headers=None,
        json_body=None,
        text=None,
        client=None,
    ):
        """
        Construct and return an SDK response object with a mocked requests.Response

        Unlike mocking of an API route, this is meant for unit testing in which we
        want to directly create the response.
        """
        from tests.common import PickleableMockResponse

        r = PickleableMockResponse(
            status, headers=headers, json_body=json_body, text=text
        )
        http_res = globus_sdk.GlobusHTTPResponse(
            r, client=client if client is not None else mock.Mock()
        )
        if response_class is not None:
            return response_class(http_res)
        return http_res

    return _make_response



from unittest import mock

import pytest

from globus_sdk.transport import (
    RequestCallerInfo,
    RequestsTransport,
    RetryCheckResult,
    RetryCheckRunner,
    RetryConfig,
    RetryContext,
)
from globus_sdk.transport.default_retry_checks import (
    DEFAULT_RETRY_CHECKS,
    check_retry_after_header,
    check_transient_error,
)


@pytest.mark.parametrize("http_status", (429, 503))
def test_retry_policy_respects_retry_after(mocksleep, http_status):
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)
    transport = RequestsTransport()
    checker = RetryCheckRunner(retry_config.checks)

    dummy_response = mock.Mock()
    dummy_response.headers = {"Retry-After": "5"}
    dummy_response.status_code = http_status
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is True
    mocksleep.assert_not_called()
    transport._retry_sleep(retry_config, ctx)
    mocksleep.assert_called_once_with(5)


@pytest.mark.parametrize("http_status", (429, 503))
def test_retry_policy_ignores_retry_after_too_high(mocksleep, http_status):
    # set explicit max sleep to confirm that the value is capped here
    retry_config = RetryConfig(max_sleep=5)
    retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)
    transport = RequestsTransport()
    checker = RetryCheckRunner(retry_config.checks)

    dummy_response = mock.Mock()
    dummy_response.headers = {"Retry-After": "20"}
    dummy_response.status_code = http_status
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is True
    mocksleep.assert_not_called()
    transport._retry_sleep(retry_config, ctx)
    mocksleep.assert_called_once_with(5)


@pytest.mark.parametrize("http_status", (429, 503))
def test_retry_policy_ignores_malformed_retry_after(mocksleep, http_status):
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)
    transport = RequestsTransport()
    checker = RetryCheckRunner(retry_config.checks)

    dummy_response = mock.Mock()
    dummy_response.headers = {"Retry-After": "not-an-integer"}
    dummy_response.status_code = http_status
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is True
    mocksleep.assert_not_called()
    transport._retry_sleep(retry_config, ctx)
    mocksleep.assert_called_once()


@pytest.mark.parametrize(
    "check_method",
    [check_retry_after_header, check_transient_error],
    ids=lambda f: f.__name__,
)
def test_default_retry_check_noop_on_exception(check_method, mocksleep):
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, exception=Exception("foo"))
    assert check_method(ctx) is RetryCheckResult.no_decision



import globus_sdk
from globus_sdk.testing import load_response


def test_register_function(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.register_function).metadata
    registration_doc = {
        "function_name": meta["function_name"],
        "function_code": meta["function_code"],
    }
    res = compute_client_v2.register_function(data=registration_doc)
    assert res.http_status == 200
    assert res.data["function_uuid"] == meta["function_id"]



from globus_sdk.testing import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_RUN

FLOW_ID = TWO_HOP_TRANSFER_RUN["flow_id"]
RUN_ID = TWO_HOP_TRANSFER_RUN["run_id"]
RESPONSES = ResponseSet(
    metadata={"run_id": RUN_ID, "flow_id": FLOW_ID},
    default=RegisteredResponse(
        service="flows",
        method="POST",
        path=f"/runs/{RUN_ID}/resume",
        json=TWO_HOP_TRANSFER_RUN,
    ),
)



import datetime

import globus_sdk
from globus_sdk.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# Globus Tutorial Collection 1
# https://app.globus.org/file-manager/collections/6c54cade-bde5-45c1-bdea-f4bd71dba2cc
SRC_COLLECTION = "6c54cade-bde5-45c1-bdea-f4bd71dba2cc"
SRC_PATH = "/home/share/godata/file1.txt"

# Globus Tutorial Collection 2
# https://app.globus.org/file-manager/collections/31ce9ba0-176d-45a5-add3-f37d233ba47d
DST_COLLECTION = "31ce9ba0-176d-45a5-add3-f37d233ba47d"
DST_PATH = "/~/example-transfer-script-destination.txt"


def make_relative_deadline(offset: datetime.timedelta) -> str:
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    deadline = now + offset
    return deadline.isoformat()


with UserApp("relative-deadline-transfer", client_id=NATIVE_CLIENT_ID) as app:
    with globus_sdk.TransferClient(app=app) as transfer_client:
        # Comment out each of these lines if the referenced collection is either
        #   (1) A guest collection or (2) high assurance.
        transfer_client.add_app_data_access_scope(SRC_COLLECTION)
        transfer_client.add_app_data_access_scope(DST_COLLECTION)

        transfer_request = globus_sdk.TransferData(
            SRC_COLLECTION,
            DST_COLLECTION,
            deadline=make_relative_deadline(datetime.timedelta(hours=1)),
        )
        transfer_request.add_item(SRC_PATH, DST_PATH)

        task = transfer_client.submit_transfer(transfer_request)
        print(f"Submitted transfer. Task ID: {task['task_id']}")



from __future__ import annotations

import typing as t
import uuid

import globus_sdk
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.scopes import GCSCollectionScopes


# NOTE: this stub class idea is inspired by the SpecificFlowScopes class stub
# it implements the same interface as the base class, so it's type-compatible
# but it raises errors at runtime -- because it can't *actually* be populated with data
class _GCSCollectionScopesClassStub(GCSCollectionScopes):
    """
    This internal stub object ensures that the type deductions for type checkers (e.g.
    mypy) on GCSCollectionClient.scopes are correct.

    Primarily, it should be possible to access the `scopes` attribute, the `user`
    scope, and the `resource_server`, but these usages should raise specific and
    informative runtime errors.

    Our types are therefore less accurate for class-var access, but more accurate for
    instance-var access.
    """

    def __init__(self) -> None:
        super().__init__("<stub>")

    def __getattribute__(self, name: str) -> t.Any:
        if name == "https":
            _raise_attr_error("https")
        if name == "data_access":
            _raise_attr_error("data_access")
        if name == "resource_server":
            _raise_attr_error("resource_server")
        return object.__getattribute__(self, name)


class GCSCollectionClient(globus_sdk.BaseClient):
    """
    A client for interacting directly with a GCS Collection.
    Typically for HTTPS upload/download via HTTPS-enabled collections.

    .. sdk-sphinx-copy-params:: BaseClient

        :param collection_id: The ID of the collection.
        :param collection_address: The URL of the collection, as might be retrieved from
            the ``https_server`` field in Globus Transfer.
    """

    scopes: GCSCollectionScopes = _GCSCollectionScopesClassStub()

    def __init__(
        self,
        collection_id: str | uuid.UUID,
        collection_address: str,
        *,
        environment: str | None = None,
        app: globus_sdk.GlobusApp | None = None,
        app_scopes: list[globus_sdk.scopes.Scope] | None = None,
        authorizer: GlobusAuthorizer | None = None,
        app_name: str | None = None,
        transport: globus_sdk.transport.RequestsTransport | None = None,
        retry_config: globus_sdk.transport.RetryConfig | None = None,
    ) -> None:
        self.collection_id = str(collection_id)
        self.scopes = GCSCollectionScopes(self.collection_id)

        if not collection_address.startswith("https://"):
            collection_address = f"https://{collection_address}"

        super().__init__(
            environment=environment,
            base_url=collection_address,
            app=app,
            app_scopes=app_scopes,
            authorizer=authorizer,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )

    @property
    def default_scope_requirements(self) -> list[globus_sdk.Scope]:
        return [self.scopes.https]


def _raise_attr_error(name: str) -> t.NoReturn:
    raise AttributeError(
        f"It is not valid to attempt to access the '{name}' attribute of the "
        "GCSCollectionClient class. "
        f"Instead, instantiate a GCSCollectionClient and access the '{name}' attribute "
        "from that instance."
    )



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID, SUBSCRIPTION_ID

ENDPOINT_CONFIG = """
display_name: My Endpoint
engine:
  type: GlobusComputeEngine
"""

DEFAULT_RESPONSE_DOC = {
    "uuid": ENDPOINT_ID,
    "name": "my-endpoint",
    "display_name": "My Endpoint",
    "multi_user": False,
    "public": False,
    "endpoint_config": ENDPOINT_CONFIG.strip(),
    "user_config_template": "",
    "user_config_schema": {},
    "description": "My endpoint description",
    "hostname": "my-hostname",
    "local_user": "user1",
    "ip_address": "140.221.112.13",
    "endpoint_version": "2.31.0",
    "sdk_version": "2.31.0",
    "subscription_uuid": SUBSCRIPTION_ID,
}

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="compute",
        path=f"/v2/endpoints/{ENDPOINT_ID}",
        method="GET",
        json=DEFAULT_RESPONSE_DOC,
    ),
)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

_kingfish = {
    "username": "kingfish@globus.org",
    "name": "Christone Ingram",
    "id": str(uuid.uuid1()),
    "email": "kingfish@globus.org",
}
_client_id = str(uuid.uuid1())
_scope = "urn:globus:auth:scope:auth.globus.org:view_identity_set profile email openid"


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token/introspect",
        method="POST",
        json={
            "active": True,
            "token_type": "Bearer",
            "scope": _scope,
            "client_id": _client_id,
            "username": _kingfish["username"],
            "name": _kingfish["name"],
            "email": _kingfish["email"],
            "exp": 1715289767,
            "iat": 1715116967,
            "nbf": 1715116967,
            "sub": _kingfish["id"],
            "aud": [
                "auth.globus.org",
                _client_id,
            ],
            "iss": "https://auth.globus.org",
        },
        metadata={
            "client_id": _client_id,
            "scope": _scope,
            **_kingfish,
        },
    )
)



import globus_sdk
from globus_sdk.testing import load_response


def test_delete_endpoint(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.delete_endpoint).metadata
    res = compute_client_v2.delete_endpoint(endpoint_id=meta["endpoint_id"])
    assert res.http_status == 200
    assert res.data == {"result": 302}



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import SUBSCRIPTION_GROUP_ID, SUBSCRIPTION_ID, SUBSCRIPTION_INFO

RESPONSES = ResponseSet(
    metadata={"group_id": SUBSCRIPTION_GROUP_ID, "subscription_id": SUBSCRIPTION_ID},
    default=RegisteredResponse(
        service="groups",
        path=f"/v2/subscription_info/{SUBSCRIPTION_ID}",
        json={
            "group_id": SUBSCRIPTION_GROUP_ID,
            "subscription_id": SUBSCRIPTION_ID,
            # this API returns restricted subscription_info
            "subscription_info": {
                k: v
                for k, v in SUBSCRIPTION_INFO.items()
                if k in ("is_high_assurance", "is_baa", "connectors")
            },
        },
    ),
)



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk import DependentScopeSpec
from globus_sdk.testing.models import RegisteredResponse, ResponseSet

SCOPE_REQUEST_ARGS = {
    "scope_id": str(uuid.uuid1()),
}


def make_request_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    request_body = {}

    for field in [
        "name",
        "description",
        "scope_suffix",
        "advertised",
        "allows_refresh_token",
        "required_domains",
    ]:
        if field in request_args and request_args[field] is not None:
            request_body[field] = request_args[field]

    if "dependent_scopes" in request_args:
        request_body["dependent_scopes"] = request_args["dependent_scopes"]

    return request_body


def make_response_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    client_id = str(uuid.uuid1())
    scope_suffix = request_args.get("scope_suffix", str(uuid.uuid4()).replace("-", ""))

    return {
        "scope_string": f"https://auth.globus.org/scopes/{client_id}/{scope_suffix}",
        "allows_refresh_token": request_args.get("allows_refresh_token", True),
        "id": request_args["scope_id"],
        "advertised": request_args.get("advertised", False),
        "required_domains": request_args.get("required_domains", []),
        "name": request_args.get("name", str(uuid.uuid4()).replace("-", "")),
        "description": request_args.get(
            "description", str(uuid.uuid4()).replace("-", "")
        ),
        "client": str(request_args.get("client_id", uuid.uuid1())),
        "dependent_scopes": [
            {
                "scope": str(ds["scope"]),
                "optional": ds["optional"],
                "requires_refresh_token": ds["requires_refresh_token"],
            }
            for ds in request_args.get("dependent_scopes", [])
        ],
    }


def register_response(
    args: t.Mapping[str, t.Any],
) -> RegisteredResponse:
    request_args = {**SCOPE_REQUEST_ARGS, **args}
    request_body = make_request_body(request_args)
    response_body = make_response_body(request_args)

    return RegisteredResponse(
        service="auth",
        method="PUT",
        path=f"/v2/api/scopes/{request_args['scope_id']}",
        json={"scope": response_body},
        metadata={
            # Test functions use 'args' to form request
            "args": request_args,
            # Test functions use 'response' to verify response
            "response": response_body,
        },
        match=[json_params_matcher({"scope": request_body})],
    )


RESPONSES = ResponseSet(
    default=register_response({}),
    name=register_response({"name": str(uuid.uuid4()).replace("-", "")}),
    description=register_response({"description": str(uuid.uuid4()).replace("-", "")}),
    scope_suffix=register_response(
        {"scope_suffix": str(uuid.uuid4()).replace("-", "")}
    ),
    no_required_domains=register_response({"required_domains": []}),
    required_domains=register_response(
        {"required_domains": ["globus.org", "uchicago.edu"]}
    ),
    no_dependent_scopes=register_response({"dependent_scopes": []}),
    dependent_scopes=register_response(
        {
            "dependent_scopes": [
                DependentScopeSpec(str(uuid.uuid1()), True, True),
                DependentScopeSpec(uuid.uuid1(), False, False),
            ],
        }
    ),
    advertised=register_response({"advertised": True}),
    not_advertised=register_response({"advertised": False}),
    allows_refresh_token=register_response({"allows_refresh_token": True}),
    disallows_refresh_token=register_response({"allows_refresh_token": False}),
)



import globus_sdk
from globus_sdk.paging import Paginator

tc = globus_sdk.TransferClient()


# return type should be IterableTransferResponse
# confirm that it is not altered by the decorator
r: globus_sdk.IterableTransferResponse = tc.task_list()
i: int = tc.task_list()  # type: ignore [assignment]
tc.task_list(limit=10, offset=100, filter="hi")
tc.task_list(limit="hi")  # type: ignore [arg-type]

# too many positional args is 'misc'
# we also will trigger arg type because `str` does not match the first kwarg (limit)
# of type `int`
tc.task_list("foo")  # type: ignore [misc,arg-type]


# the same basic should hold for a Paginator.wrap'ed variant
# but the return type should be a paginator of responses
paginated_call = Paginator.wrap(tc.task_list)
paged_r: Paginator[globus_sdk.IterableTransferResponse] = paginated_call()
i = paginated_call()  # type: ignore [assignment]
paginated_call(limit=10, offset=100, filter="hi")
paginated_call(limit="hi")  # type: ignore [arg-type]

# see note above on non-paginated case for why these ignores are correct
paginated_call("foo")  # type: ignore [misc,arg-type]



import uuid

import pytest

from globus_sdk.scopes import Scope


def test_scope_with_dependency_leaves_original_unchanged():
    s1 = Scope(uuid.uuid1().hex)
    s2 = Scope("s2")
    s3 = s1.with_dependency(s2)

    assert s1.scope_string == s3.scope_string
    assert len(s1.dependencies) == 0
    assert len(s3.dependencies) == 1


def test_scope_with_dependencies_leaves_original_unchanged():
    s1 = Scope(uuid.uuid1().hex)
    s2 = Scope("s2")
    s3 = Scope("s3")
    s4 = s1.with_dependencies((s2, s3))

    assert s1.scope_string == s4.scope_string
    assert len(s1.dependencies) == 0
    assert len(s4.dependencies) == 2


def test_scope_with_optional_leaves_original_unchanged():
    s1 = Scope(uuid.uuid1().hex)
    s2 = s1.with_optional(True)
    s3 = s2.with_optional(False)

    assert s1.scope_string == s2.scope_string == s3.scope_string
    assert len(s1.dependencies) == 0
    assert len(s2.dependencies) == 0
    assert len(s3.dependencies) == 0

    assert not s1.optional
    assert s2.optional
    assert not s3.optional


def test_scope_with_string_dependency_gets_typeerror():
    s = Scope("x")
    with pytest.raises(
        TypeError,
        match=r"Scope\.with_dependency\(\) takes a Scope as its input\. Got: 'str'",
    ):
        s.with_dependency("y")


def test_scope_with_string_dependencies_gets_typeerror():
    s = Scope("x")
    with pytest.raises(
        TypeError,
        match=(
            r"Scope\.with_dependencies\(\) takes an iterable of Scopes as its input\. "
            "At position 0, got: 'str'"
        ),
    ):
        s.with_dependencies(["y"])


def test_scope_with_mixed_dependencies_gets_typeerror():
    s = Scope("x")
    with pytest.raises(
        TypeError,
        match=(
            r"Scope\.with_dependencies\(\) takes an iterable of Scopes as its input\. "
            "At position 1, got: 'str'"
        ),
    ):
        s.with_dependencies([Scope("y"), "z"])



import typing as t
import uuid

import globus_sdk

ac = globus_sdk.AuthLoginClient("someclientid")

# base case
url: str = ac.oauth2_get_authorize_url()


def generate_strs() -> t.Iterator[str]:
    yield "foo"
    yield "bar"


# with session_required_{identities,single_domain,policies} as strings
url = ac.oauth2_get_authorize_url(session_required_identities="foo")
url = ac.oauth2_get_authorize_url(session_required_single_domain="foo")
url = ac.oauth2_get_authorize_url(session_required_policies="foo")
url = ac.oauth2_get_authorize_url(session_message="foo")

# or any iterable of strings
url = ac.oauth2_get_authorize_url(session_required_identities=generate_strs())
url = ac.oauth2_get_authorize_url(session_required_single_domain=generate_strs())
url = ac.oauth2_get_authorize_url(session_required_policies=generate_strs())

# these two support UUIDs
url = ac.oauth2_get_authorize_url(session_required_identities=uuid.uuid4())
url = ac.oauth2_get_authorize_url(session_required_policies=uuid.uuid4())

# now the negative tests, starting with the fact that domain can't take a UUID
url = ac.oauth2_get_authorize_url(
    session_required_single_domain=uuid.uuid4()  # type: ignore[arg-type]
)
url = ac.oauth2_get_authorize_url(
    session_message=uuid.uuid4(),  # type: ignore[arg-type]
)

# integers and other non-str data are not acceptable
url = ac.oauth2_get_authorize_url(
    session_required_identities=1  # type: ignore[arg-type]
)
url = ac.oauth2_get_authorize_url(
    session_required_single_domain=1  # type: ignore[arg-type]
)
url = ac.oauth2_get_authorize_url(session_required_policies=1)  # type: ignore[arg-type]
url = ac.oauth2_get_authorize_url(session_message=1)  # type: ignore[arg-type]



from responses.matchers import query_param_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

API_VERSION = "1.23.0"
ALL_RESPONSE_DOC = {
    "api": API_VERSION,
    "min_sdk_version": "1.0.0a6",
    "min_endpoint_version": "1.0.0a0",
    "git_sha": "80b2ef87bc546b3b386cf2e1d372f4be50f10bc4",
}

RESPONSES = ResponseSet(
    metadata={"api_version": API_VERSION},
    default=RegisteredResponse(
        service="compute",
        path="/v2/version",
        method="GET",
        json=API_VERSION,  # type: ignore[arg-type]
    ),
    all=RegisteredResponse(
        service="compute",
        path="/v2/version",
        method="GET",
        json=ALL_RESPONSE_DOC,
        match=[query_param_matcher(params={"service": "all"})],
    ),
)



import pytest


def test_extdoclink_renders_simple(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. extdoclink:: Create Demuddler
            :ref: demuddler/create
        """,
    )

    assert etree.tag == "document"

    paragraph = etree.find("paragraph")
    assert paragraph is not None
    text_parts = list(paragraph.itertext())
    assert "See " in text_parts[0]
    assert "API documentation for details" in text_parts[-1]

    link = paragraph.find("reference")
    assert link is not None
    assert link.text == "Create Demuddler"
    uri = link.get("refuri")
    assert uri is not None
    assert uri == "https://docs.globus.org/api/demuddler/create"


@pytest.mark.parametrize(
    "service, base_url",
    (
        ("groups", "https://groups.api.globus.org/redoc#operation"),
        ("gcs", "https://docs.globus.org/globus-connect-server/v5/api"),
        ("flows", "https://globusonline.github.io/globus-flows#tag"),
        ("compute", "https://compute.api.globus.org/redoc#tag"),
    ),
)
def test_extdoclink_renders_with_custom_service(sphinx_runner, service, base_url):
    etree = sphinx_runner.to_etree(
        f"""\
        .. extdoclink:: Battle Cry
            :ref: theTick/SPOOOON
            :service: {service}
        """,
    )

    assert etree.tag == "document"

    paragraph = etree.find("paragraph")
    assert paragraph is not None
    text_parts = list(paragraph.itertext())
    assert "See " in text_parts[0]
    assert "API documentation for details" in text_parts[-1]

    link = paragraph.find("reference")
    assert link is not None
    assert link.text == "Battle Cry"
    uri = link.get("refuri")
    assert uri is not None
    assert uri == f"{base_url}/theTick/SPOOOON"


def test_extdoclink_renders_with_custom_base_url(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. extdoclink:: Battle Cry
            :ref: SPOOOON
            :base_url: https://docs.globus.org/90sNostalgia/TheCity
        """,
    )

    assert etree.tag == "document"

    paragraph = etree.find("paragraph")
    assert paragraph is not None
    link = paragraph.find("reference")
    assert link is not None
    assert link.text == "Battle Cry"
    uri = link.get("refuri")
    assert uri is not None
    assert uri == "https://docs.globus.org/90sNostalgia/TheCity/SPOOOON"



from __future__ import annotations

import abc
import contextlib
import copy
import logging
import sys
import types
import typing as t
import uuid
from json import JSONDecodeError

from requests import Response

from globus_sdk import (
    AuthClient,
    AuthLoginClient,
    GlobusSDKUsageError,
    IDTokenDecoder,
)
from globus_sdk._internal.type_definitions import Closable
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.gare import GARE, GlobusAuthorizationParameters, to_gare
from globus_sdk.scopes import AuthScopes, Scope, ScopeParser
from globus_sdk.token_storage import (
    ScopeRequirementsValidator,
    TokenStorage,
    TokenValidationError,
    ValidatingTokenStorage,
)
from globus_sdk.transport import (
    RetryCheck,
    RetryCheckFlags,
    RetryCheckResult,
    RetryContext,
    set_retry_check_flags,
)

from .authorizer_factory import AuthorizerFactory
from .config import DEFAULT_CONFIG, KNOWN_TOKEN_STORAGES, GlobusAppConfig
from .protocols import TokenStorageProvider

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

log = logging.getLogger(__name__)


class GlobusApp(metaclass=abc.ABCMeta):
    """
    The abstract base class for managing authentication across services.

    A single ``GlobusApp`` may be bound to many service clients, providing each them
    with dynamically updated authorization tokens. The app is responsible for ensuring
    these tokens are up-to-date and validly scoped; including initiating login flows to
    acquire new tokens when necessary.

    See :class:`UserApp` to oversee interactions with a human.

    See :class:`ClientApp` to oversee interactions with a service account.

    .. warning::

        GlobusApp is **not** thread safe.

    :ivar ValidatingTokenStorage token_storage: The interface used by the app to store,
        retrieve, and validate Globus Auth-issued tokens.
    :ivar dict[str, list[Scope]] scope_requirements: A copy of the app's aggregate scope
        requirements. Modifying the returned dict will not affect the app's internal
        store. To add scope requirements, instead use the :meth:`add_scope_requirements`
        method.
    """

    _login_client: AuthLoginClient
    # a bool is used to track whether or not the AuthorizerFactory is ready for use
    # this allows code during init call into otherwise unsafe codepaths in the app,
    # namely those which manipulate scope requirements
    _authorizer_factory: AuthorizerFactory[GlobusAuthorizer]
    _token_storage: TokenStorage
    token_storage: ValidatingTokenStorage

    def __init__(
        self,
        app_name: str = "Unnamed Globus App",
        *,
        login_client: AuthLoginClient | None = None,
        client_id: uuid.UUID | str | None = None,
        client_secret: str | None = None,
        scope_requirements: (
            t.Mapping[str, str | Scope | t.Iterable[str | Scope]] | None
        ) = None,
        config: GlobusAppConfig = DEFAULT_CONFIG,
    ) -> None:
        self.app_name = app_name
        self.config = config
        self._token_validation_error_handling_enabled = True
        self._resources_to_close: list[Closable] = []

        self.client_id, self._login_client = self._resolve_client_info(
            app_name=self.app_name,
            config=self.config,
            client_id=client_id,
            client_secret=client_secret,
            login_client=login_client,
        )
        self._scope_requirements = self._resolve_scope_requirements(scope_requirements)

        # create the inner token storage object, and pick up on whether or not this
        # call handled creation or got a value from the outside
        self._token_storage, token_storage_created_here = self._resolve_token_storage(
            app_name=self.app_name,
            client_id=self.client_id,
            config=self.config,
        )

        # create a consent client for token validation
        # this client won't be ready for immediate use, but will have the app attached
        # at the end of init
        consent_client = AuthClient(environment=config.environment)
        self._resources_to_close.append(consent_client)

        # create the requisite token storage for the app, with validation based on
        # the provided parameters
        # if the token storage was created by the app, the validating wrapper is added
        # to the resources to close when closed
        self.token_storage = self._initialize_validating_token_storage(
            token_storage=self._token_storage,
            consent_client=consent_client,
            scope_requirements=self._scope_requirements,
        )
        if token_storage_created_here:
            self._resources_to_close.append(self.token_storage)

        # setup an ID Token Decoder based on config; build one if it was not provided
        self._id_token_decoder = self._initialize_id_token_decoder(
            app_name=self.app_name, config=self.config, login_client=self._login_client
        )

        # initialize our authorizer factory
        self._initialize_authorizer_factory()
        self._authorizer_factory_initialized = True

        # finally, attach the app to the internal consent client
        # this needs to wait until the very end of the app initialization process so
        # that the authorizer factory is all ready to accept the client
        # registering its scope requirements
        #
        # additionally, this will ensure that openid scope requirement is always
        # registered (it's required for token identity validation).
        consent_client.attach_globus_app(self, app_scopes=[AuthScopes.openid])

    def _resolve_scope_requirements(
        self,
        scope_requirements: (
            t.Mapping[str, str | Scope | t.Iterable[str | Scope]] | None
        ),
    ) -> dict[str, list[Scope]]:
        if scope_requirements is None:
            return {}

        return {
            resource_server: list(self._iter_scopes(scopes))
            for resource_server, scopes in scope_requirements.items()
        }

    def _resolve_client_info(
        self,
        app_name: str,
        config: GlobusAppConfig,
        login_client: AuthLoginClient | None,
        client_id: uuid.UUID | str | None,
        client_secret: str | None,
    ) -> tuple[uuid.UUID | str, AuthLoginClient]:
        """
        Extracts a client_id and login_client from GlobusApp initialization parameters,
        validating that the parameters were provided correctly.

        Depending on which parameters were provided, this method will either:
            1.  Create a new login client from the supplied credentials.
                *   The actual initialization is performed by the subclass using the
                    abstract method: _initialize_login_client``.
            2.  Extract the client_id from a supplied login_client.

        If a new client is created here, it is also added to the set of resources to
        close when this app is closed.

        :returns: tuple of client_id and login_client
        :raises: GlobusSDKUsageError if a single client ID or login client could not be
            definitively resolved.
        """
        if login_client and client_id:
            msg = "Mutually exclusive parameters: client_id and login_client."
            raise GlobusSDKUsageError(msg)

        if login_client:
            # User provided an explicit login client, extract the client_id.
            if login_client.client_id is None:
                msg = "A GlobusApp login_client must have a discoverable client_id."
                raise GlobusSDKUsageError(msg)
            if login_client.environment != config.environment:
                raise GlobusSDKUsageError(
                    "[Environment Mismatch] The login_client's environment "
                    f"({login_client.environment}) does not match the GlobusApp's "
                    f"configured environment ({config.environment})."
                )

            return login_client.client_id, login_client

        elif client_id:
            # User provided an explicit client_id, construct a login client
            login_client = self._initialize_login_client(
                app_name, config, client_id, client_secret
            )
            self._resources_to_close.append(login_client)
            return client_id, login_client

        else:
            raise GlobusSDKUsageError(
                "Could not set up a globus login client. One of client_id or "
                "login_client is required."
            )

    @abc.abstractmethod
    def _initialize_login_client(
        self,
        app_name: str,
        config: GlobusAppConfig,
        client_id: uuid.UUID | str,
        client_secret: str | None,
    ) -> AuthLoginClient:
        """
        Initializes and returns an AuthLoginClient to be used in authorization requests.
        """

    def _initialize_validating_token_storage(
        self,
        token_storage: TokenStorage,
        consent_client: AuthClient,
        scope_requirements: t.Mapping[str, t.Sequence[Scope]],
    ) -> ValidatingTokenStorage:
        """
        Initializes the validating token storage for the app.
        """
        validating_token_storage = ValidatingTokenStorage(token_storage)

        # construct ValidatingTokenStorage around the TokenStorage and
        # our initial scope requirements
        scope_validator = ScopeRequirementsValidator(scope_requirements, consent_client)

        # use validators to enforce invariants about scopes
        validating_token_storage.validators.append(scope_validator)

        return validating_token_storage

    def _resolve_token_storage(
        self, app_name: str, client_id: uuid.UUID | str, config: GlobusAppConfig
    ) -> tuple[TokenStorage, bool]:
        """
        Resolve the raw token storage to be used by the app, and whether or not it was
        created explicitly here.

        The storage may be:
            1.  A TokenStorage instance provided by the user, which we use directly.
            2.  A TokenStorageProvider, which we use to get a TokenStorage.
            3.  A string value, which we map onto supported TokenStorage types.

        And in the case of (1), the bool is false. In the case of (2) and (3) it is
        true.

        :returns: TokenStorage instance to be used by the app, and whether or not this
            function created the storage.
        :raises: GlobusSDKUsageError if the provided token_storage value is unsupported.
        """
        token_storage = config.token_storage
        # TODO - make namespace configurable
        namespace = "DEFAULT"
        if isinstance(token_storage, TokenStorage):
            return (token_storage, False)

        elif isinstance(token_storage, TokenStorageProvider):
            return (
                token_storage.for_globus_app(
                    app_name=app_name,
                    config=config,
                    client_id=client_id,
                    namespace=namespace,
                ),
                True,
            )

        elif token_storage in KNOWN_TOKEN_STORAGES:
            provider = KNOWN_TOKEN_STORAGES[token_storage]
            return (
                provider.for_globus_app(
                    app_name=app_name,
                    config=config,
                    client_id=client_id,
                    namespace=namespace,
                ),
                True,
            )

        raise GlobusSDKUsageError(
            f"Unsupported token_storage value: {token_storage}. Must be a "
            f"TokenStorage, TokenStorageProvider, or a supported string value."
        )

    def _initialize_id_token_decoder(
        self, *, app_name: str, config: GlobusAppConfig, login_client: AuthLoginClient
    ) -> IDTokenDecoder:
        """
        Create an IDTokenDecoder or use the one provided via config, and set it on
        the token storage adapters.

        It is only set on inner storage if the decoder was not already set, so a
        non-null value won't be overwritten.

        This must run near the end of app initialization, when the `_token_storage`
        (inner) and `token_storage` (validating storage, outer) storages have both
        been initialized.
        """
        if isinstance(self.config.id_token_decoder, IDTokenDecoder):
            id_token_decoder: IDTokenDecoder = self.config.id_token_decoder
        else:
            id_token_decoder = self.config.id_token_decoder.for_globus_app(
                app_name=app_name,
                config=config,
                login_client=login_client,
            )
        if self._token_storage.id_token_decoder is None:
            self._token_storage.id_token_decoder = id_token_decoder
        self.token_storage.id_token_decoder = id_token_decoder

        return id_token_decoder

    @abc.abstractmethod
    def _initialize_authorizer_factory(self) -> None:
        """
        Initializes self._authorizer_factory to be used for generating authorizers to
        authorize requests.
        """

    def login(
        self,
        *,
        auth_params: GlobusAuthorizationParameters | None = None,
        force: bool = False,
    ) -> None:
        """
        Log a user or client into the app, if needed, storing the resulting tokens.

        A login flow will be performed if any of the following are true:
            * The kwarg ``auth_params`` is provided.
            * The kwarg ``force`` is set to True.
            * The method ``self.login_required()`` evaluates to True.

        :param auth_params: An optional set of authorization parameters to establish
            requirements and controls for the login flow.
        :param force: If True, perform a login flow even if one does not appear to
            be necessary.
        """
        if auth_params or force or self.login_required():
            self._run_login_flow(auth_params)

    def login_required(self) -> bool:
        """
        Determine if a login flow will be required to interact with resource servers
        under the current scope requirements.

        This will return false if any of the following are true:
            * Access tokens have never been issued.
            * Access tokens have been issued but have insufficient scopes.
            * Access tokens have expired and wouldn't be resolved with refresh tokens.

        :returns: True if a login flow appears to be required, False otherwise.
        """
        for resource_server in self._scope_requirements.keys():
            try:
                with self._disabled_token_validation_error_handler():
                    self.get_authorizer(resource_server)
            except TokenValidationError:
                return True
        return False

    def logout(self) -> None:
        """
        Log the current user or client out of the app.

        This will remove and revoke all tokens stored for the current app user.
        """
        # Revoke all tokens, removing them from the underlying token storage
        inner_token_storage = self.token_storage.token_storage
        for resource_server in self._scope_requirements.keys():
            token_data = inner_token_storage.get_token_data(resource_server)
            if token_data:
                self._login_client.oauth2_revoke_token(token_data.access_token)
                if token_data.refresh_token:
                    self._login_client.oauth2_revoke_token(token_data.refresh_token)
                inner_token_storage.remove_token_data(resource_server)

        # Invalidate any cached authorizers
        self._authorizer_factory.clear_cache()

    def close(self) -> None:
        """
        Close all resources currently held by the app.
        This does not trigger a logout.
        """
        for resource in self._resources_to_close:
            log.debug(
                f"closing resource of type {type(resource).__name__} "
                f"for {type(self).__name__}(app_name={self.app_name!r})"
            )
            resource.close()

    # apps can act as context managers, and such usage calls close()
    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        self.close()

    @abc.abstractmethod
    def _run_login_flow(
        self, auth_params: GlobusAuthorizationParameters | None = None
    ) -> None:
        """
        Run an authorization flow to get new tokens which are stored and available
        for the next authorizer gotten by get_authorizer.

        :param auth_params: An optional set of authorization parameters to establish
            requirements and controls for the login flow.
        """

    def _auth_params_with_required_scopes(
        self,
        auth_params: GlobusAuthorizationParameters | None = None,
    ) -> GlobusAuthorizationParameters:
        """
        Either make a new GlobusAuthorizationParameters with this app's required scopes
        or combine this app's required scopes with given auth_params.
        """
        required_scopes = []
        for scope_list in self._scope_requirements.values():
            required_scopes.extend(scope_list)

        if not auth_params:
            auth_params = GlobusAuthorizationParameters()
        parsed_required_scopes = []
        for s in auth_params.required_scopes or []:
            parsed_required_scopes.extend(ScopeParser.parse(s))

        # merge scopes for deduplication to minimize url request length
        # this is useful even if there weren't any auth_param scope requirements
        # as the app's scope_requirements can have duplicates
        combined_scopes = ScopeParser.merge_scopes(
            required_scopes, parsed_required_scopes
        )
        auth_params.required_scopes = [str(s) for s in combined_scopes]

        return auth_params

    @contextlib.contextmanager
    def _disabled_token_validation_error_handler(self) -> t.Iterator[None]:
        """
        Context manager to disable token validation error handling (as a default)
        for the duration of the context.
        """
        # Record the starting value so we can reset it after the context ends.
        initial_val = self._token_validation_error_handling_enabled

        self._token_validation_error_handling_enabled = False
        try:
            yield
        finally:
            self._token_validation_error_handling_enabled = initial_val

    def add_scope_requirements(
        self, scope_requirements: t.Mapping[str, str | Scope | t.Iterable[str | Scope]]
    ) -> None:
        """
        Add given scope requirements to the app's scope requirements. Any duplicate
        requirements will be deduplicated with existing requirements.

        :param scope_requirements: a dict of Scopes indexed by resource server
            that will be added to this app's scope requirements
        """
        for resource_server, scopes in scope_requirements.items():
            curr = self._scope_requirements.setdefault(resource_server, [])
            curr.extend(self._iter_scopes(scopes))

        self._authorizer_factory.clear_cache(*scope_requirements.keys())

    def get_authorizer(self, resource_server: str) -> GlobusAuthorizer:
        """
        Get a ``GlobusAuthorizer`` for a resource server.

        This method will be called by service clients while making HTTP requests.

        :param resource_server: The resource server for which the requested Authorizer
            should provide authorization headers.
        """
        error_handling_enabled = self._token_validation_error_handling_enabled

        try:
            # Disable token validation error handling for nested calls.
            # This will ultimately ensure that the error handler is only called once
            # by the root `get_authorizer` invocation.
            with self._disabled_token_validation_error_handler():
                return self._authorizer_factory.get_authorizer(resource_server)
        except TokenValidationError as e:
            if error_handling_enabled and self.config.token_validation_error_handler:
                # Dispatch to the configured error handler if one is set then retry.
                self.config.token_validation_error_handler(self, e)
                return self._authorizer_factory.get_authorizer(resource_server)
            raise e

    def get_client_retry_checks(self) -> t.Sequence[RetryCheck]:
        if not self.config.auto_redrive_gares:
            return []
        else:
            return [_RedriveGlobusAppGARE(self)]

    @property
    def scope_requirements(self) -> dict[str, list[Scope]]:
        """
        Access a copy of app's aggregate scope requirements.

        Modifying the returned dict will not affect the app's scope requirements.
        To add scope requirements, use ``GlobusApp.add_scope_requirements()``.

        .. note::

            Users may observe that Globus Auth (``'auth.globus.org'``) is always
            present, and always maps to the ``openid`` scope, even when the user has
            not added this scope. This mapping is expected, as the ``openid`` scope is
            needed internally for the functionality provided by ``GlobusApp``.
        """
        # Scopes are mutable objects so we return a deepcopy
        return copy.deepcopy(self._scope_requirements)

    def _iter_scopes(
        self, scopes: str | Scope | t.Iterable[str | Scope]
    ) -> t.Iterator[Scope]:
        """Normalize scopes in various formats to an iterator of Scope objects."""
        if isinstance(scopes, str):
            yield from ScopeParser.parse(scopes)
        elif isinstance(scopes, Scope):
            yield scopes
        else:
            for item in scopes:
                if isinstance(item, str):
                    yield from ScopeParser.parse(item)
                else:
                    yield item


@set_retry_check_flags(RetryCheckFlags.RUN_ONCE)
class _RedriveGlobusAppGARE:
    """
    A client.transport RetryCheck specific to GlobusApps.

    Upon receiving a GARE-parsable 403 response, this check will initiate a login,
    refresh the authorizer, and dispatch a retry.
    """

    def __init__(self, app: GlobusApp) -> None:
        self.app = app

    def __call__(self, ctx: RetryContext) -> RetryCheckResult:
        if (resource_server := ctx.caller_info.resource_server) is None:
            return RetryCheckResult.no_decision

        elif (gare := self._load_response_gare(ctx.response)) is None:
            return RetryCheckResult.no_decision

        log.debug("Intercepted re-drivable GARE; initiating app login.")
        self.app.login(auth_params=gare.authorization_parameters)
        updated_authorizer = self.app.get_authorizer(resource_server)
        log.debug("Acquired updated authorizer after GARE login; retrying.")

        ctx.caller_info.authorizer = updated_authorizer
        return RetryCheckResult.do_retry

    @staticmethod
    def _load_response_gare(response: Response | None) -> GARE | None:
        """Return a parsed GARE from a 403 response or None if not possible."""
        if response is None or response.status_code != 403:
            return None
        try:
            decoded_body = response.json()
        except JSONDecodeError:
            return None
        else:
            return to_gare(decoded_body)



from __future__ import annotations

import typing as t

from globus_sdk._internal.guards import validators
from globus_sdk._internal.serializable import Serializable


class GlobusAuthorizationParameters(Serializable):
    """
    Data class containing authorization parameters that can be passed during
    an authentication flow to control how the user will authenticate.

    When used with a GARE this represents the additional authorization
    parameters needed in order to complete a request that had insufficient
    authorization state.

    :ivar session_message: A message to be displayed to the user.
    :vartype session_message: str, optional

    :ivar session_required_identities: A list of identities required for the
        session.
    :vartype session_required_identities: list of str, optional

    :ivar session_required_policies: A list of policies required for the
        session.
    :vartype session_required_policies: list of str, optional

    :ivar session_required_single_domain: A list of domains required for the
        session.
    :vartype session_required_single_domain: list of str, optional

    :ivar session_required_mfa: Whether MFA is required for the session.
    :vartype session_required_mfa: bool, optional

    :ivar required_scopes: A list of scopes for which consent is required.
    :vartype required_scopes: list of str, optional

    :ivar prompt: The OIDC 'prompt' parameter, for which Globus Auth currently supports
        the values 'login' and 'none'.
    :vartype prompt: str, optional

    :ivar extra: A dictionary of additional fields that were provided. May
        be used for forward/backward compatibility.
    :vartype extra: dict
    """

    def __init__(
        self,
        *,
        session_message: str | None = None,
        session_required_identities: list[str] | None = None,
        session_required_policies: list[str] | None = None,
        session_required_single_domain: list[str] | None = None,
        session_required_mfa: bool | None = None,
        required_scopes: list[str] | None = None,
        prompt: str | None = None,
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        self.session_message = validators.opt_str("session_message", session_message)
        self.session_required_identities = validators.opt_str_list(
            "session_required_identities", session_required_identities
        )
        self.session_required_policies = validators.opt_str_list(
            "session_required_policies", session_required_policies
        )
        self.session_required_single_domain = validators.opt_str_list(
            "session_required_single_domain", session_required_single_domain
        )
        self.session_required_mfa = validators.opt_bool(
            "session_required_mfa", session_required_mfa
        )
        self.required_scopes = validators.opt_str_list(
            "required_scopes", required_scopes
        )
        self.prompt = validators.opt_str("prompt", prompt)
        self.extra = extra or {}

    def __repr__(self) -> str:
        extra_repr = ""
        if self.extra:
            extra_repr = ", extra=..."
        attrs = [
            f"{name}={getattr(self, name)!r}"
            for name in (
                "session_message",
                "session_required_identities",
                "session_required_policies",
                "session_required_single_domain",
                "session_required_mfa",
                "required_scopes",
                "prompt",
            )
        ]
        return "GlobusAuthorizationParameters(" + ", ".join(attrs) + extra_repr + ")"


class GARE(Serializable):
    """
    Represents a Globus Auth Requirements Error.

    A Globus Auth Requirements Error is a class of error that is returned by Globus
    services to indicate that additional authorization is required in order to complete
    a request and contains information that can be used to request the appropriate
    authorization.

    :ivar code: The error code for this error.
    :vartype code: str

    :ivar authorization_parameters: The authorization parameters for this error.
    :vartype authorization_parameters: GlobusAuthorizationParameters

    :ivar extra: A dictionary of additional fields that were provided. May
        be used for forward/backward compatibility.
    :vartype extra: dict
    """

    def __init__(
        self,
        code: str,
        authorization_parameters: dict[str, t.Any] | GlobusAuthorizationParameters,
        *,
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        self.code = validators.str_("code", code)
        self.authorization_parameters = validators.instance_or_dict(
            "authorization_parameters",
            authorization_parameters,
            GlobusAuthorizationParameters,
        )
        self.extra = extra or {}

    def __repr__(self) -> str:
        extra_repr = ""
        if self.extra:
            extra_repr = ", extra=..."
        return (
            f"GARE(code={self.code!r}, "
            f"authorization_parameters={self.authorization_parameters!r}"
            f"{extra_repr})"
        )



import json
import urllib.parse

import pytest

from globus_sdk import MISSING
from globus_sdk.testing import RegisteredResponse, get_last_request, load_response


@pytest.fixture(autouse=True)
def setup_mock_responses():
    load_response(
        RegisteredResponse(
            path="https://foo.api.globus.org/bar",
            json={"foo": "bar"},
        )
    )
    load_response(
        RegisteredResponse(
            path="https://foo.api.globus.org/bar",
            method="POST",
            json={"foo": "bar"},
        )
    )


def test_query_params_can_filter_missing(client):
    res = client.get("/bar", query_params={"foo": "bar", "baz": MISSING})
    assert res.http_status == 200
    req = get_last_request()
    assert req.params == {"foo": "bar"}


def test_headers_can_filter_missing(client):
    res = client.get("/bar", headers={"foo": "bar", "baz": MISSING})
    assert res.http_status == 200
    req = get_last_request()
    assert req.headers["foo"] == "bar"
    assert "baz" not in req.headers


def test_json_body_can_filter_missing(client):
    res = client.post("/bar", data={"foo": "bar", "baz": MISSING})
    assert res.http_status == 200
    req = get_last_request()
    sent = json.loads(req.body)
    assert sent == {"foo": "bar"}


def test_form_body_can_filter_missing(client):
    res = client.post("/bar", data={"foo": "bar", "baz": MISSING}, encoding="form")
    assert res.http_status == 200
    req = get_last_request()
    sent = urllib.parse.parse_qs(req.body)
    assert sent == {"foo": ["bar"]}



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_DOC, TWO_HOP_TRANSFER_FLOW_ID

_DELETED_DOC = {
    "DELETED": True,
    "administered_by": [],
    "api_version": "1.0",
    "created_at": TWO_HOP_TRANSFER_FLOW_DOC["created_at"],
    "created_by": TWO_HOP_TRANSFER_FLOW_DOC["created_by"],
    "definition": TWO_HOP_TRANSFER_FLOW_DOC["definition"],
    "deleted_at": "2022-10-20T16:44:59.126641+00:00",
    "description": "",
    "flow_administrators": [],
    "flow_owner": TWO_HOP_TRANSFER_FLOW_DOC["flow_owner"],
    "flow_starters": [],
    "flow_viewers": [],
    "globus_auth_scope": TWO_HOP_TRANSFER_FLOW_DOC["globus_auth_scope"],
    "globus_auth_username": TWO_HOP_TRANSFER_FLOW_DOC["globus_auth_username"],
    "id": TWO_HOP_TRANSFER_FLOW_ID,
    "input_schema": {},
    "keywords": [],
    "log_supported": True,
    "principal_urn": TWO_HOP_TRANSFER_FLOW_DOC["principal_urn"],
    "runnable_by": [],
    "subtitle": "",
    "synchronous": False,
    "title": TWO_HOP_TRANSFER_FLOW_DOC["title"],
    "types": ["Action"],
    "updated_at": "2022-10-20T16:44:59.021201+00:00",
    "user_role": "flow_owner",
    "visible_to": [],
}

RESPONSES = ResponseSet(
    metadata={"flow_id": TWO_HOP_TRANSFER_FLOW_ID},
    default=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}",
        method="DELETE",
        json=_DELETED_DOC,
    ),
)



from __future__ import annotations

import pytest

from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "case_name",
    (
        "template_id_str",
        "template_id_uuid",
        "name",
    ),
)
def test_create_native_app_instance(
    auth_client,
    case_name: str,
):
    meta = load_response(
        auth_client.create_native_app_instance, case=case_name
    ).metadata

    res = auth_client.create_native_app_instance(**meta["args"])
    for k, v in meta["response"].items():
        assert res["client"][k] == v



from __future__ import annotations

import typing as t
import uuid

from .base import TokenStorage
from .token_data import TokenStorageData

if t.TYPE_CHECKING:
    from globus_sdk.globus_app import GlobusAppConfig


class MemoryTokenStorage(TokenStorage):
    """
    A token storage which holds tokens in-memory.
    All token data is lost when the process exits.

    See :class:`TokenStorage` for common interface details.

    :param namespace: A unique string for partitioning token data (Default: "DEFAULT").
    """

    def __init__(self, *, namespace: str = "DEFAULT") -> None:
        self._tokens: dict[str, dict[str, t.Any]] = {}
        super().__init__(namespace=namespace)

    @classmethod
    def for_globus_app(
        cls,
        # pylint: disable=unused-argument
        client_id: uuid.UUID | str,
        app_name: str,
        config: GlobusAppConfig,
        # pylint: enable=unused-argument
        namespace: str,
    ) -> MemoryTokenStorage:
        return cls(namespace=namespace)

    def store_token_data_by_resource_server(
        self, token_data_by_resource_server: t.Mapping[str, TokenStorageData]
    ) -> None:
        if self.namespace not in self._tokens:
            self._tokens[self.namespace] = {}

        for resource_server, token_data in token_data_by_resource_server.items():
            self._tokens[self.namespace][resource_server] = token_data.to_dict()

    def get_token_data_by_resource_server(self) -> dict[str, TokenStorageData]:
        ret = {}
        dicts_by_resource_server = self._tokens.get(self.namespace, {})
        for resource_server, token_data_dict in dicts_by_resource_server.items():
            ret[resource_server] = TokenStorageData.from_dict(token_data_dict)
        return ret

    def remove_token_data(self, resource_server: str) -> bool:
        popped = self._tokens.get(self.namespace, {}).pop(resource_server, None)
        return popped is not None



class LocalServerLoginError(Exception):
    """An error raised during a LocalServerLoginFlowManager's run."""


class LocalServerEnvironmentalLoginError(LocalServerLoginError):
    """
    Error raised when a local server login flow fails to start due to incompatible
    environment conditions (e.g., a remote session or text-only browser).
    """



from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk import client, exc, response
from globus_sdk._internal import guards
from globus_sdk.scopes import (
    GCSCollectionScopes,
    Scope,
    SpecificFlowScopes,
    TimersScopes,
    TransferScopes,
)

from .data import FlowTimer, TimerJob, TransferTimer
from .errors import TimersAPIError

log = logging.getLogger(__name__)


class TimersClient(client.BaseClient):
    r"""
    Client for the Globus Timers API.

    .. sdk-sphinx-copy-params:: BaseClient

    .. automethodlist:: globus_sdk.TimersClient
    """

    error_class = TimersAPIError
    service_name = "timer"
    scopes = TimersScopes
    default_scope_requirements = [TimersScopes.timer]

    def add_app_transfer_data_access_scope(
        self, collection_ids: uuid.UUID | str | t.Iterable[uuid.UUID | str]
    ) -> TimersClient:
        """
        Add a dependent ``data_access`` scope for one or more given ``collection_ids``
        to this client's ``GlobusApp``, under the Transfer ``all`` scope.
        Useful for preventing ``ConsentRequired`` errors when creating timers
        that use Globus Connect Server mapped collection(s) as the source or
        destination.

        .. warning::

            This method must only be used on ``collection_ids`` for non-High-Assurance
            GCS Mapped Collections.

            Use on other collection types, e.g., on GCP Mapped Collections or any form
            of Guest Collection, will result in "Unknown Scope" errors during the login
            flow.

        Returns ``self`` for chaining.

        Raises ``GlobusSDKUsageError`` if this client was not initialized with an app.

        :param collection_ids: a collection ID or an iterable of IDs.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    app = UserApp("myapp", client_id=NATIVE_APP_CLIENT_ID)
                    client = TimersClient(app=app).add_app_transfer_data_access_scope(COLLECTION_ID)

                    transfer_data = TransferData(COLLECTION_ID, COLLECTION_ID)
                    transfer_data.add_item("/staging/", "/active/")

                    daily_timer = TransferTimer(
                        name="test_timer", schedule=RecurringTimerSchedule(24 * 60 * 60), body=transfer_data
                    )

                    client.create_timer(daily_timer)
        """  # noqa: E501
        if isinstance(collection_ids, (str, uuid.UUID)):
            guards.validators.uuidlike("collection_ids", collection_ids)
            # wrap the collection_ids input in a list for consistent iteration below
            collection_ids_ = [collection_ids]
        else:
            # copy to a list so that ephemeral iterables can be iterated multiple times
            collection_ids_ = list(collection_ids)
            for i, c in enumerate(collection_ids_):
                guards.validators.uuidlike(f"collection_ids[{i}]", c)

        dependencies: list[Scope] = []
        for coll_id in collection_ids_:
            data_access_scope = GCSCollectionScopes(
                str(coll_id)
            ).data_access.with_optional(True)
            dependencies.append(data_access_scope)
        transfer_scope = TransferScopes.all.with_dependencies(dependencies)

        timers_scope = TimersScopes.timer.with_dependency(transfer_scope)
        self.add_app_scope(timers_scope)
        return self

    def add_app_flow_user_scope(
        self, flow_ids: uuid.UUID | str | t.Iterable[uuid.UUID | str]
    ) -> TimersClient:
        """
        Add a dependent flow ``user`` scope for one or more given ``flow_ids``
        to this client's ``GlobusApp``, under the Timers ``timer`` scope.
        Needed to prevent ``ConsentRequired`` errors when creating flow timers.

        Returns ``self`` for chaining.

        Raises ``GlobusSDKUsageError`` if this client was not initialized with an app.

        :param flow_ids: a flow ID or an iterable of IDs.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    app = UserApp("myapp", client_id=NATIVE_APP_CLIENT_ID)
                    client = TimersClient(app=app)
                    client.add_app_flow_user_scope(FLOW_ID)

                    flow_timer = FlowTimer(
                        name="my flow timer",
                        flow_id=FLOW_ID,
                        schedule=RecurringTimerSchedule(24 * 60 * 60),
                        body={
                            "body": {
                                "input_key": "input_value",
                            },
                        },
                    )

                    client.create_timer(flow_timer)
        """  # noqa: E501
        if isinstance(flow_ids, (str, uuid.UUID)):
            guards.validators.uuidlike("flow_ids", flow_ids)
            # wrap the flow_ids input in a list for consistent iteration below
            flow_ids_ = [flow_ids]
        else:
            # copy to a list so that ephemeral iterables can be iterated multiple times
            flow_ids_ = list(flow_ids)
            for i, c in enumerate(flow_ids_):
                guards.validators.uuidlike(f"flow_ids[{i}]", c)

        dependencies: list[Scope] = [
            SpecificFlowScopes(flow_id).user for flow_id in flow_ids_
        ]
        timers_scope = TimersScopes.timer.with_dependencies(dependencies)
        self.add_app_scope(timers_scope)
        return self

    def list_jobs(
        self, *, query_params: dict[str, t.Any] | None = None
    ) -> response.GlobusHTTPResponse:
        """
        ``GET /jobs/``

        :param query_params: additional parameters to pass as query params

        **Examples**

        >>> timer_client = globus_sdk.TimersClient(...)
        >>> jobs = timer_client.list_jobs()
        """
        log.debug(f"TimersClient.list_jobs({query_params})")
        return self.get("/jobs/", query_params=query_params)

    def get_job(
        self,
        job_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        ``GET /jobs/<job_id>``

        :param job_id: the ID of the timer ("job")
        :param query_params: additional parameters to pass as query params

        **Examples**

        >>> timer_client = globus_sdk.TimersClient(...)
        >>> job = timer_client.get_job(job_id)
        >>> assert job["job_id"] == job_id
        """
        log.debug(f"TimersClient.get_job({job_id})")
        return self.get(f"/jobs/{job_id}", query_params=query_params)

    def create_timer(
        self, timer: dict[str, t.Any] | TransferTimer | FlowTimer
    ) -> response.GlobusHTTPResponse:
        """
        :param timer: a document defining the new timer

        A ``TransferTimer`` object can be constructed from a ``TransferData`` object,
        which is the recommended way to create a timer for data transfers.

        **Examples**

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> transfer_data = TransferData(...)
                    >>> timers_client = globus_sdk.TimersClient(...)
                    >>> create_doc = globus_sdk.TransferTimer(
                    ...     name="my-timer",
                    ...     schedule={"type": "recurring", "interval": 1800},
                    ...     body=transfer_data,
                    ... )
                    >>> response = timers_client.create_timer(timer=create_doc)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: timer.create_timer

            .. tab-item:: API Info

                ``POST /v2/timer``
        """
        if isinstance(timer, TimerJob):
            raise exc.GlobusSDKUsageError(
                "Cannot pass a TimerJob to create_timer(). "
                "Create a TransferTimer instead."
            )
        log.debug("TimersClient.create_timer(...)")
        return self.post("/v2/timer", data={"timer": timer})

    def create_job(
        self, data: dict[str, t.Any] | TimerJob
    ) -> response.GlobusHTTPResponse:
        """
        ``POST /jobs/``

        :param data: a timer document used to create the new timer ("job")

        **Examples**

        >>> from datetime import datetime, timedelta, timezone
        >>> callback_url = ...
        >>> data = ...
        >>> timer_client = globus_sdk.TimersClient(...)
        >>> job = TimerJob(
        ...     callback_url,
        ...     data,
        ...     datetime.now(tz=timezone.utc).replace(tzinfo=None),
        ...     timedelta(days=14),
        ...     name="my-timer-job"
        ... )
        >>> timer_result = timers_client.create_job(job)
        """
        if isinstance(data, TransferTimer):
            raise exc.GlobusSDKUsageError(
                "Cannot pass a TransferTimer to create_job(). Use create_timer() "
                "instead."
            )
        log.debug(f"TimersClient.create_job({data})")
        return self.post("/jobs/", data=data)

    def update_job(
        self, job_id: uuid.UUID | str, data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        ``PATCH /jobs/<job_id>``

        :param job_id: the ID of the timer ("job")
        :param data: a partial timer document used to update the job

        **Examples**

        >>> timer_client = globus_sdk.TimersClient(...)
        >>> timer_client.update_job(job_id, {"name": "new name}"})
        """
        log.debug(f"TimersClient.update_job({job_id}, {data})")
        return self.patch(f"/jobs/{job_id}", data=data)

    def delete_job(
        self,
        job_id: uuid.UUID | str,
    ) -> response.GlobusHTTPResponse:
        """
        ``DELETE /jobs/<job_id>``

        :param job_id: the ID of the timer ("job")

        **Examples**

        >>> timer_client = globus_sdk.TimersClient(...)
        >>> timer_client.delete_job(job_id)
        """
        log.debug(f"TimersClient.delete_job({job_id})")
        return self.delete(f"/jobs/{job_id}")

    def pause_job(
        self,
        job_id: uuid.UUID | str,
    ) -> response.GlobusHTTPResponse:
        """
        Make a timer job inactive, preventing it from running until it is resumed.

        :param job_id: The ID of the timer to pause

        **Examples**

        >>> timer_client = globus_sdk.TimersClient(...)
        >>> timer_client.pause_job(job_id)
        """
        log.debug(f"TimersClient.pause_job({job_id})")
        return self.post(f"/jobs/{job_id}/pause")

    def resume_job(
        self,
        job_id: uuid.UUID | str,
        *,
        update_credentials: bool | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Resume an inactive timer job, optionally replacing credentials to resolve
        issues with insufficient authorization.

        :param job_id: The ID of the timer to resume
        :param update_credentials: When true, replace the credentials for the timer
            using the credentials for this resume call. This can be used to resolve
            authorization errors (such as session and consent errors), but it also
            could introduce session and consent errors, if the credentials being used
            to resume lack some necessary properties of the credentials they're
            replacing. If not supplied, the Timers service will determine whether to
            replace credentials according to the reason why the timer job
            became inactive.

        **Examples**

        >>> timer_client = globus_sdk.TimersClient(...)
        >>> timer_client.resume_job(job_id)
        """
        log.debug(f"TimersClient.resume_job({job_id})")
        data = {}
        if update_credentials is not None:
            data["update_credentials"] = update_credentials
        return self.post(f"/jobs/{job_id}/resume", data=data)



"""
Tooling for an extremely simple lazy-import system, based on inspection of
pyi files.

Given a base module name (used for lookup and error messages) and the name
of a pyi file, we can use the pyi file to lookup import locations.

i.e. Given

  foo.py
  foo.pyi
  bar.py

then if `foo.pyi` has an import `from .bar import BarType`, it is possible to
*read* `foo.pyi` at runtime and use that information to load
`BarType` from `bar`.


Rationale
---------

Why use this type of lazy importer?

The major goals are:
- type information is still available to type checkers (and other tools like IDEs)
- we minimize manual duplication of names
- we minimize specialized knowledge needed to update our package exports
- the lazy importer itself is easy to maintain and update

``.pyi`` type stubs are designed for use with type checkers. The data format choice here
is "regular type stubs". We treat it as runtime data, which is a bit unusual, but we
are guaranteed that `mypy` and other tools can understand it without issue.

We reduce duplication with this technique, and the content in the ``.pyi`` file should
be easy for most developers to read and modify.
"""

from __future__ import annotations

import ast
import sys
import typing as t


def load_all_tuple(modname: str, pyi_filename: str) -> tuple[str, ...]:
    """
    Load the __all__ tuple from a ``.pyi`` file.

    This should run before the getattr and dir implementations are defined, as those use
    the runtime ``__all__`` tuple.

    :param modname: The name of the module doing the load. Usually ``__name__``.
    :param pyi_filename: The name of the ``pyi`` file relative to ``modname``.
        ``importlib.resources`` will use both of these fields to load the ``pyi``
        data, so the file must be in package metadata.
    """
    return _ParsedPYIData.load(modname, pyi_filename).all_names


def default_getattr_implementation(
    modname: str, pyi_filename: str
) -> t.Callable[[str], t.Any]:
    """
    Build an implementation of module ``__getattr__`` given the module name and
    the pyi file which will drive lazy imports.

    :param modname: The name of the module where ``__getattr__`` is being added.
        Usually ``__name__``.
    :param pyi_filename: The name of the ``pyi`` file relative to ``modname``.
        ``importlib.resources`` will use both of these fields to load the ``pyi``
        data, so the file must be in package metadata.
    """
    module_object = sys.modules[modname]
    all_tuple = module_object.__all__

    def getattr_implementation(name: str) -> t.Any:
        if name in all_tuple:
            value = load_attr(modname, pyi_filename, name)
            setattr(module_object, name, value)
            return value

        raise AttributeError(f"module {modname} has no attribute {name}")

    return getattr_implementation


def default_dir_implementation(modname: str) -> t.Callable[[], list[str]]:
    """
    Build an implementation of module ``__dir__`` given the module name.

    :param modname: The name of the module where ``__dir__`` is being added.
        Usually ``__name__``.
    """
    # dir(globus_sdk) should include everything exported in __all__
    # as well as some explicitly selected attributes from the default dir() output
    # on a module
    #
    # see also:
    # https://discuss.python.org/t/how-to-properly-extend-standard-dir-search-with-module-level-dir/4202
    module_object = sys.modules[modname]
    all_tuple = module_object.__all__

    def dir_implementation() -> list[str]:
        return list(all_tuple) + [
            # __all__ itself can be inspected
            "__all__",
            # useful to figure out where a package is installed
            "__file__",
            "__path__",
        ]

    return dir_implementation


def load_attr(modname: str, pyi_filename: str, attrname: str) -> t.Any:
    """
    Execute an import of a single attribute in the manner that it was declared in a
    ``.pyi`` file.

    The import in the pyi data is expected to be a `from x import y` statement.
    Only the specific attribute will be imported, even if the pyi declares multiple
    imports from the same module.

    :param modname: The name of the module importing the attribute.
        Usually ``__name__``.
    :param pyi_filename: The name of the ``pyi`` file relative to ``modname``.
        ``importlib.resources`` will use both of these fields to load the ``pyi``
        data, so the file must be in package metadata.
    :param attrname: The name of the attribute to load.
    """
    import importlib

    attr_source = find_source_module(modname, pyi_filename, attrname)
    attr_source_mod = importlib.import_module(attr_source, modname)
    return getattr(attr_source_mod, attrname)


def find_source_module(modname: str, pyi_filename: str, attrname: str) -> str:
    """
    Find the source module which provides an attribute, based on a declared import in a
    ``.pyi`` file.

    The ``.pyi`` data will be parsed as AST and scanned for an appropriate import.

    :param modname: The name of the module importing the attribute.
        Usually ``__name__``.
    :param pyi_filename: The name of the ``pyi`` file relative to ``modname``.
        ``importlib.resources`` will use both of these fields to load the ``pyi``
        data, so the file must be in package metadata.
    :param attrname: The name of the attribute to load.
    """
    parsed = _ParsedPYIData.load(modname, pyi_filename)
    return parsed.module_for_attr(attrname)


class _ParsedPYIData:
    _CACHE: dict[tuple[str, str], _ParsedPYIData] = {}

    @classmethod
    def load(cls, module_name: str, pyi_filename: str) -> _ParsedPYIData:
        if (module_name, pyi_filename) not in cls._CACHE:
            cls._CACHE[(module_name, pyi_filename)] = cls(module_name, pyi_filename)
        return cls._CACHE[(module_name, pyi_filename)]

    def __init__(self, module_name: str, pyi_filename: str) -> None:
        self.module_name = module_name
        self.pyi_filename = pyi_filename
        self._ast = _parse_pyi_ast(module_name, pyi_filename)

        self._import_attr_map: dict[str, str] = {}
        self._all_names: list[str] = []

        self._load()

    @property
    def all_names(self) -> tuple[str, ...]:
        return tuple(self._all_names)

    def module_for_attr(self, attrname: str) -> str:
        if attrname not in self._import_attr_map:
            raise LookupError(
                f"Could not find import of '{attrname}' in '{self.module_name}'."
            )
        return self._import_attr_map[attrname]

    def _load(self) -> None:
        for statement in self._ast.body:
            self._load_statement(statement)

    def _load_statement(self, statement: ast.AST) -> None:
        if isinstance(statement, ast.ImportFrom):
            # type ignore the possibility of 'import_from.module == None'
            # as it's not possible from parsed code
            module_name = (  # type: ignore[operator]
                "." * statement.level
            ) + statement.module

            for alias in statement.names:
                attr_name = alias.name if alias.asname is None else alias.asname
                self._import_attr_map[attr_name] = module_name
        elif isinstance(statement, ast.Assign):
            if len(statement.targets) != 1:
                return
            target = statement.targets[0]
            if not isinstance(target, ast.Name):
                return
            if target.id != "__all__":
                return
            if not isinstance(statement.value, ast.Tuple):
                raise ValueError(
                    f"While reading '__all__' for '{self.module_name}' from "
                    f"'{self.pyi_filename}', '__all__' was not a tuple "
                )
            for element in statement.value.elts:
                if not isinstance(element, ast.Constant) or not isinstance(
                    element.value, str
                ):
                    continue
                self._all_names.append(element.value)


def _parse_pyi_ast(anchor_module_name: str, pyi_filename: str) -> ast.Module:
    import importlib.resources

    source = (
        importlib.resources.files(anchor_module_name)  # pylint: disable=no-member
        .joinpath(pyi_filename)
        .read_bytes()
    )
    return ast.parse(source)



import json

import pytest
from responses import matchers

from globus_sdk import MISSING, FlowsAPIError
from globus_sdk.testing import get_last_request, load_response
from globus_sdk.testing.models import RegisteredResponse


@pytest.mark.parametrize("subscription_id", [MISSING, None, "dummy_subscription_id"])
def test_create_flow(flows_client, subscription_id):
    metadata = load_response(flows_client.create_flow).metadata

    resp = flows_client.create_flow(
        **metadata["params"], subscription_id=subscription_id
    )
    assert resp.data["title"] == "Multi Step Transfer"

    last_req = get_last_request()
    req_body = json.loads(last_req.body)
    if subscription_id is not MISSING:
        assert req_body["subscription_id"] == subscription_id
    else:
        assert "subscription_id" not in req_body


@pytest.mark.parametrize("value", [MISSING, [], ["dummy_value"]])
@pytest.mark.parametrize("key", ["run_managers", "run_monitors"])
def test_create_flow_run_role_serialization(flows_client, key, value):

    request_body = {
        "title": "Multi Step Transfer",
        "definition": {
            "StartAt": "Transfer1",
            "States": {
                "Transfer1": {"Type": "Pass", "End": True},
            },
        },
        "input_schema": {},
    }

    if value is not MISSING:
        request_body[key] = value

    load_response(
        RegisteredResponse(
            service="flows",
            path="/flows",
            method="POST",
            json=request_body,
            match=[
                matchers.json_params_matcher(
                    params={
                        "title": request_body["title"],
                        "definition": request_body["definition"],
                        "input_schema": request_body["input_schema"],
                    },
                    strict_match=False,
                )
            ],
        )
    )

    flows_client.create_flow(**request_body)

    last_req = get_last_request()
    req_body = json.loads(last_req.body)

    if value is MISSING:
        assert key not in req_body
    else:
        assert req_body[key] == value


def test_create_flow_error_parsing(flows_client):
    metadata = load_response(
        flows_client.create_flow, case="bad_admin_principal_error"
    ).metadata
    with pytest.raises(FlowsAPIError) as excinfo:
        flows_client.create_flow(**metadata["params"])
    err = excinfo.value
    assert err.code == "UNPROCESSABLE_ENTITY"
    assert err.messages == [
        (
            'Unrecognized principal string: "ae341a98-d274-11e5-b888-dbae3a8ba545". '
            'Allowed principal types in role "FlowAdministrator": '
            "[<AuthGroupUrn>, <AuthIdentityUrn>]"
        ),
        (
            'Unrecognized principal string: "4fab4345-6d20-43a0-9a25-16b2e3bbe765". '
            'Allowed principal types in role "FlowAdministrator": '
            "[<AuthGroupUrn>, <AuthIdentityUrn>]"
        ),
    ]


def test_get_flow(flows_client):
    meta = load_response(flows_client.get_flow).metadata
    resp = flows_client.get_flow(meta["flow_id"])
    assert resp.data["title"] == meta["title"]


def test_update_flow(flows_client):
    meta = load_response(flows_client.update_flow).metadata
    resp = flows_client.update_flow(meta["flow_id"], **meta["params"])
    for k, v in meta["params"].items():
        assert k in resp
        assert resp[k] == v


@pytest.mark.parametrize("value", [MISSING, [], ["dummy_value"]])
@pytest.mark.parametrize("key", ["run_managers", "run_monitors"])
def test_update_flow_run_role_serialization(flows_client, key, value):
    metadata = load_response(flows_client.update_flow).metadata
    params = {**metadata["params"], key: value}

    flows_client.update_flow(metadata["flow_id"], **params)

    last_req = get_last_request()
    req_body = json.loads(last_req.body)

    if value is MISSING:
        assert key not in req_body
    else:
        assert req_body[key] == value


def test_delete_flow(flows_client):
    metadata = load_response(flows_client.delete_flow).metadata

    resp = flows_client.delete_flow(metadata["flow_id"])
    assert resp.data["title"] == "Multi Step Transfer"
    assert resp.data["DELETED"] is True
